import re
from enum import IntEnum
from typing import Callable, List

from sqlfmt.node import Node
from sqlfmt.tokens import TokenType


class OperatorPrecedence(IntEnum):
    DOUBLE_COLON = 0
    AS = 1
    SQUARE_BRACKETS = 2
    OTHER_TIGHT = 3
    EXPONENT = 4
    MULTIPLICATION = 5
    ADDITION = 6
    OTHER = 7
    MEMBERSHIP = 8
    COMPARATORS = 9
    PRESENCE = 10
    BOOL_NOT = 11
    BOOL_AND = 12
    BOOL_OR = 13
    ON = 14

    @staticmethod
    def tiers() -> List["OperatorPrecedence"]:
        return [
            OperatorPrecedence.OTHER_TIGHT,
            OperatorPrecedence.EXPONENT,
            OperatorPrecedence.MULTIPLICATION,
            OperatorPrecedence.OTHER,
            OperatorPrecedence.COMPARATORS,
            OperatorPrecedence.BOOL_NOT,
            OperatorPrecedence.ON,
        ]

    @classmethod
    def from_node(cls, node: Node) -> "OperatorPrecedence":
        assert node.is_operator, (
            f"Internal error! {node} is not an operator. Please open an issue"
        )
        return cls._function_lookup(token_type=node.token.type)(node)

    @classmethod
    def _function_lookup(
        cls,
        token_type: TokenType,
    ) -> Callable[[Node], "OperatorPrecedence"]:
        mapping = {
            TokenType.DOUBLE_COLON: lambda x: OperatorPrecedence.DOUBLE_COLON,
            TokenType.BRACKET_OPEN: lambda x: OperatorPrecedence.SQUARE_BRACKETS,
            TokenType.ON: lambda x: OperatorPrecedence.ON,
            TokenType.STAR: lambda x: OperatorPrecedence.MULTIPLICATION,
            TokenType.BOOLEAN_OPERATOR: cls._from_boolean,
            TokenType.OPERATOR: cls._from_operator,
            TokenType.WORD_OPERATOR: cls._from_word_operator,
        }
        return mapping.get(token_type, lambda x: OperatorPrecedence.OTHER)

    @staticmethod
    def _from_operator(node: Node) -> "OperatorPrecedence":
        value_mapping = {
            "+": OperatorPrecedence.ADDITION,
            "-": OperatorPrecedence.ADDITION,
            "/": OperatorPrecedence.MULTIPLICATION,
            "%": OperatorPrecedence.MULTIPLICATION,
            "%%": OperatorPrecedence.MULTIPLICATION,
            "^": OperatorPrecedence.EXPONENT,
            "=": OperatorPrecedence.COMPARATORS,
            "==": OperatorPrecedence.COMPARATORS,
            "!=": OperatorPrecedence.COMPARATORS,
            "<>": OperatorPrecedence.COMPARATORS,
            "<=": OperatorPrecedence.COMPARATORS,
            ">=": OperatorPrecedence.COMPARATORS,
            ">": OperatorPrecedence.COMPARATORS,
            "<": OperatorPrecedence.COMPARATORS,
            "~": OperatorPrecedence.MEMBERSHIP,
            "~*": OperatorPrecedence.MEMBERSHIP,
            "!~": OperatorPrecedence.MEMBERSHIP,
            "!~*": OperatorPrecedence.MEMBERSHIP,
            "~~": OperatorPrecedence.MEMBERSHIP,
            "~~*": OperatorPrecedence.MEMBERSHIP,
            "!~~": OperatorPrecedence.MEMBERSHIP,
            "!~~*": OperatorPrecedence.MEMBERSHIP,
        }
        return value_mapping.get(node.value, OperatorPrecedence.OTHER)

    @staticmethod
    def _from_word_operator(node: Node) -> "OperatorPrecedence":
        mapping = [
            (OperatorPrecedence.AS, [r"as"]),
            (
                OperatorPrecedence.OTHER_TIGHT,
                [
                    r"exclude",
                    r"replace",
                    r"except",
                    r"over",
                    r"within\s+group",
                    r"filter",
                    r"using",
                ],
            ),
            (
                OperatorPrecedence.MEMBERSHIP,
                [
                    r"(not\s+)?between",
                    r"(not\s+)?in",
                    r"(not\s+)?i?like(\s+any)?",
                    r"(not\s+)?similar\s+to",
                    r"(not\s+)?rlike",
                    r"(not\s+)?regexp",
                ],
            ),
            (OperatorPrecedence.PRESENCE, [r"is(\s+not)?", r"isnull", r"notnull"]),
        ]

        for precedence, pattern_list in mapping:
            programs = [
                re.compile(f"{pattern}$", flags=re.IGNORECASE)
                for pattern in pattern_list
            ]
            if any([prog.match(node.value) for prog in programs]):
                return precedence
        else:
            return OperatorPrecedence.OTHER

    @staticmethod
    def _from_boolean(node: Node) -> "OperatorPrecedence":
        if node.is_the_and_after_the_between_operator:
            return OperatorPrecedence.OTHER_TIGHT

        value_mapping = {
            "and": OperatorPrecedence.BOOL_AND,
            "or": OperatorPrecedence.BOOL_OR,
            "not": OperatorPrecedence.BOOL_NOT,
        }
        return value_mapping.get(node.value, OperatorPrecedence.BOOL_AND)



from collections.abc import Iterable, Mapping
from typing import Any, Generic, Iterator, MutableMapping, TypeVar, overload

from _typeshed import SupportsWrite

__all__ = [
    "tqdm",
]

_T = TypeVar("_T")

class Comparable:
    pass

class tqdm(Generic[_T], Iterable[_T], Comparable):
    @overload
    def __init__(
        self,
        iterable: Iterable[_T],
        desc: str | None = ...,
        total: float | None = ...,
        leave: bool | None = ...,
        file: SupportsWrite[str] | None = ...,
        ncols: int | None = ...,
        mininterval: float = ...,
        maxinterval: float = ...,
        miniters: float | None = ...,
        ascii: bool | str | None = ...,
        disable: bool | None = ...,
        unit: str = ...,
        unit_scale: bool | float = ...,
        dynamic_ncols: bool = ...,
        smoothing: float = ...,
        bar_format: str | None = ...,
        initial: float = ...,
        position: int | None = ...,
        postfix: Mapping[str, object] | str | None = ...,
        unit_divisor: float = ...,
        write_bytes: bool | None = ...,
        lock_args: tuple[bool | None, float | None] | tuple[bool | None] | None = ...,
        nrows: int | None = ...,
        colour: str | None = ...,
        delay: float | None = ...,
        gui: bool = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self,
        iterable: None = ...,
        desc: str | None = ...,
        total: float | None = ...,
        leave: bool | None = ...,
        file: SupportsWrite[str] | None = ...,
        ncols: int | None = ...,
        mininterval: float = ...,
        maxinterval: float = ...,
        miniters: float | None = ...,
        ascii: bool | str | None = ...,
        disable: bool | None = ...,
        unit: str = ...,
        unit_scale: bool | float = ...,
        dynamic_ncols: bool = ...,
        smoothing: float = ...,
        bar_format: str | None = ...,
        initial: float = ...,
        position: int | None = ...,
        postfix: Mapping[str, object] | str | None = ...,
        unit_divisor: float = ...,
        write_bytes: bool | None = ...,
        lock_args: tuple[bool | None, float | None] | tuple[bool | None] | None = ...,
        nrows: int | None = ...,
        colour: str | None = ...,
        delay: float | None = ...,
        gui: bool = ...,
        **kwargs: Any,
    ) -> None: ...
    def update(self, n: float | None = ...) -> bool | None: ...
    def close(self) -> None: ...
    def clear(self, nolock: bool = ...) -> None: ...
    @property
    def format_dict(self) -> MutableMapping[str, Any]: ...
    def __iter__(self) -> Iterator[_T]: ...



import itertools
from collections import Counter
from typing import List

import pytest

from sqlfmt.rule import Rule
from sqlfmt.rules import (
    CLONE,
    CORE,
    FUNCTION,
    GRANT,
    JINJA,
    MAIN,
    UNSUPPORTED,
    WAREHOUSE,
)

ALL_RULESETS = [CLONE, CORE, FUNCTION, GRANT, JINJA, MAIN, WAREHOUSE, UNSUPPORTED]


def get_rule(ruleset: List[Rule], rule_name: str) -> Rule:
    """
    Return the rule from ruleset that matches rule_name
    """
    matching_rules = filter(lambda rule: rule.name == rule_name, ruleset)
    try:
        return next(matching_rules)
    except StopIteration as e:
        raise ValueError(f"No rule '{rule_name}' in ruleset '{ruleset}'") from e


@pytest.mark.parametrize(
    "ruleset,rule_name,value",
    [
        (CORE, "fmt_off", "-- fmt: off"),
        (CORE, "fmt_off", "--FMT: off"),
        (CORE, "fmt_off", "# fmt: off"),
        (CORE, "fmt_on", "--fmt: ON"),
        (CORE, "jinja_start", "{%"),
        (CORE, "jinja_end", "}}"),
        (CORE, "quoted_name", "`my_backticked_field_name`"),
        (CORE, "quoted_name", "'my_quoted_literal'"),
        (CORE, "quoted_name", '"my_quoted_field_name"'),
        (CORE, "quoted_name", '"""triple " quotes!"""'),
        (CORE, "quoted_name", "'''triple '' singles!\\'''"),
        (CORE, "quoted_name", "$$ dollar \ndelimited'$$"),
        (CORE, "quoted_name", "$$ SELECT $1, CAST($1 AS text) || ' is text' $$"),
        (CORE, "quoted_name", "$label$ dollar delimited with label$label$"),
        (CORE, "quoted_name", "'single quote with '' doubled escape'"),
        (CORE, "quoted_name", "'single quote \\' c escape'"),
        (CORE, "quoted_name", '"double quote with "" doubled escape"'),
        (CORE, "quoted_name", '"double quote \\" c escape"'),
        (CORE, "quoted_name", "`backtick with \\` c escape`"),
        (CORE, "quoted_name", 'r"bq raw string"'),
        (CORE, "quoted_name", "U&'pg unicode string'"),
        (CORE, "quoted_name", 'rb"""bq raw bytes string"""'),
        (CORE, "quoted_name", '@"my!QUOTED!stage!name"'),
        (CORE, "comment", "-- my comment"),
        (CORE, "comment", "--no-space comment"),
        (CORE, "comment", "# mysql-style # comments"),
        (CORE, "comment", "#nospace"),
        (CORE, "comment", "// why is this a comment??!"),
        (CORE, "comment", "//whywhywhy??"),
        (CORE, "comment_start", "/*"),
        (CORE, "comment_end", "*/"),
        (MAIN, "statement_start", "case"),
        (MAIN, "statement_end", "END"),
        (CORE, "star", "*"),
        (CORE, "spark_int_literals", "145y"),
        (CORE, "spark_int_literals", "-145s"),
        (CORE, "spark_int_literals", "5l"),
        (CORE, "number", "145.8"),
        (CORE, "number", "-.58"),
        (CORE, "number", "+145.8"),
        (CORE, "number", "+.58"),
        (CORE, "number", "1e9"),
        (CORE, "number", "1e-9"),
        (CORE, "number", "1.55e-9"),
        (CORE, "number", "145.8bd"),
        (CORE, "number", "-.58d"),
        (CORE, "number", "+145.8f"),
        (CORE, "number", "+.58bd"),
        (CORE, "number", "1e9d"),
        (CORE, "number", "1e-9f"),
        (CORE, "number", "1.55e-9bd"),
        (CORE, "number", "0xFF"),
        (CORE, "number", "0Xff"),
        (CORE, "number", "0o1234"),
        (CORE, "number", "0b010101010"),
        (CORE, "number", "1_000_000"),
        (CORE, "number", "3.1415_9265"),
        (CORE, "bracket_open", "["),
        (CORE, "bracket_close", ")"),
        (CORE, "double_colon", "::"),
        (CORE, "walrus", ":="),
        (CORE, "colon", ":"),
        (CORE, "semicolon", ";"),
        (CORE, "operator", "+"),
        (CORE, "operator", "-"),
        (CORE, "operator", "/"),
        (CORE, "operator", "%"),
        (CORE, "angle_bracket_close", ">"),
        (CORE, "operator", ">"),
        (CORE, "operator", "%%"),
        (CORE, "operator", ">="),
        (CORE, "operator", "<>"),
        (CORE, "operator", "||"),
        (CORE, "operator", "=>"),
        (CORE, "operator", "<=>"),
        (CORE, "operator", "||/"),
        (CORE, "operator", "|/"),
        (CORE, "operator", ">>"),
        (CORE, "operator", "<<"),
        (CORE, "operator", "!"),
        (CORE, "operator", "!="),
        # posix like/ not like
        (CORE, "operator", "~"),
        (CORE, "operator", "!~"),
        (CORE, "operator", "~*"),
        (CORE, "operator", "!~*"),
        # postgres (not) like/ilike operators
        # like posix but doubles the tilde
        # https://github.com/tconbeer/sqlfmt/issues/576
        (CORE, "operator", "~~"),
        (CORE, "operator", "!~~"),
        (CORE, "operator", "~~*"),
        (CORE, "operator", "!~~*"),
        # postgresql geo operators
        # see: https://www.postgresql.org/docs/current/functions-geometry.html
        (CORE, "operator", "@-@"),
        (CORE, "operator", "@@"),
        (CORE, "pg_operator", "##"),
        (CORE, "operator", "<->"),
        (CORE, "operator", "<@"),
        (CORE, "operator", "@>"),
        (CORE, "operator", "?&"),
        (CORE, "operator", "?|"),
        (CORE, "operator", "&&"),
        (CORE, "operator", "&<"),
        (CORE, "operator", "&>"),
        (CORE, "operator", "<<|"),
        (CORE, "operator", "|>>"),
        (CORE, "operator", "&<|"),
        (CORE, "operator", "|&>"),
        (CORE, "operator", "<^"),
        (CORE, "operator", ">^"),
        (CORE, "operator", "?"),
        (CORE, "operator", "?#"),
        (CORE, "operator", "?-"),
        (CORE, "operator", "?|"),
        (CORE, "operator", "?-|"),
        (CORE, "operator", "?||"),
        (CORE, "operator", "~="),
        # network operators
        # see https://www.postgresql.org/docs/current/functions-net.html
        (CORE, "operator", "<<="),
        (CORE, "operator", ">>="),
        # json operators
        # see https://www.postgresql.org/docs/current/functions-json.html
        (CORE, "operator", "->"),
        (CORE, "operator", "->>"),
        (CORE, "pg_operator", "#>"),
        (CORE, "pg_operator", "#>>"),
        (CORE, "operator", "-|-"),  # range adjacency
        (MAIN, "word_operator", "is"),
        (MAIN, "word_operator", "is not"),
        (MAIN, "word_operator", "is distinct from"),
        (MAIN, "word_operator", "is not distinct from"),
        (MAIN, "word_operator", "in"),
        (MAIN, "word_operator", "not in"),
        (MAIN, "word_operator", "not\n\nin"),
        (MAIN, "word_operator", "interval"),
        (MAIN, "word_operator", "like"),
        (MAIN, "word_operator", "not like"),
        (MAIN, "word_operator", "ilike"),
        (MAIN, "word_operator", "not ilike"),
        (MAIN, "word_operator", "rlike"),
        (MAIN, "word_operator", "not rlike"),
        (MAIN, "word_operator", "like any"),
        (MAIN, "word_operator", "not like any"),
        (MAIN, "word_operator", "like all"),
        (MAIN, "word_operator", "not like all"),
        (MAIN, "word_operator", "some"),
        (MAIN, "word_operator", "exists"),
        (MAIN, "word_operator", "not exists"),
        (MAIN, "word_operator", "grouping sets"),
        (MAIN, "word_operator", "cube"),
        (MAIN, "word_operator", "rollup"),
        (MAIN, "word_operator", "over"),
        (MAIN, "word_operator", "within group"),
        (MAIN, "word_operator", "filter"),
        (MAIN, "word_operator", "as"),
        (MAIN, "word_operator", "tablesample"),
        (MAIN, "word_operator", "pivot"),
        (MAIN, "word_operator", "unpivot"),
        (MAIN, "on", "on"),
        (MAIN, "boolean_operator", "AND"),
        (CORE, "comma", ","),
        (CORE, "dot", "."),
        (MAIN, "unterm_keyword", "select DISTINCT"),
        (MAIN, "unterm_keyword", "select"),
        (MAIN, "unterm_keyword", "select\n\t    distinct"),
        (MAIN, "unterm_keyword", "select top 25"),
        (MAIN, "unterm_keyword", "select all"),
        (MAIN, "unterm_keyword", "natural\t    full outer join"),
        (MAIN, "unterm_keyword", "left join"),
        (MAIN, "unterm_keyword", "cross join"),
        (MAIN, "unterm_keyword", "positional join"),
        (MAIN, "unterm_keyword", "asof join"),
        (MAIN, "unterm_keyword", "asof left join"),
        (MAIN, "unterm_keyword", "left asof join"),
        (MAIN, "unterm_keyword", "asof right\n outer\n join"),
        (MAIN, "unterm_keyword", "semi join"),
        (MAIN, "unterm_keyword", "left anti join"),
        (MAIN, "unterm_keyword", "right anti join"),
        (MAIN, "unterm_keyword", "anti join"),
        (MAIN, "unterm_keyword", "any join"),
        (MAIN, "unterm_keyword", "left any join"),
        (MAIN, "unterm_keyword", "any left join"),
        (MAIN, "unterm_keyword", "all join"),
        (MAIN, "unterm_keyword", "array join"),
        (MAIN, "unterm_keyword", "left array join"),
        (MAIN, "unterm_keyword", "paste join"),
        (MAIN, "unterm_keyword", "join"),
        (MAIN, "unterm_keyword", "values"),
        (MAIN, "unterm_keyword", "cluster by"),
        (MAIN, "unterm_keyword", "sort\nby"),
        (MAIN, "unterm_keyword", "distribute\t by"),
        (MAIN, "unterm_keyword", "lateral view"),
        (MAIN, "unterm_keyword", "lateral view outer"),
        (MAIN, "unterm_keyword", "delete from"),
        (MAIN, "set_operator", "union"),
        (MAIN, "set_operator", "union all"),
        (MAIN, "set_operator", "union distinct"),
        (MAIN, "set_operator", "union by name"),
        (MAIN, "set_operator", "union all by name"),
        (MAIN, "set_operator", "LEFT OUTER UNION ALL BY NAME"),
        (MAIN, "set_operator", "FULL OUTER UNION ALL BY NAME"),
        (MAIN, "set_operator", "INTERSECT DISTINCT BY NAME"),
        (MAIN, "set_operator", "INTERSECT DISTINCT STRICT CORRESPONDING"),
        (MAIN, "set_operator", "except corresponding by"),
        (MAIN, "set_operator", "intersect"),
        (MAIN, "set_operator", "minus"),
        (MAIN, "set_operator", "except"),
        (CORE, "bracket_open", "array<"),
        (CORE, "bracket_open", "map<"),
        (CORE, "bracket_open", "table\n<"),
        (CORE, "bracket_open", "struct<"),
        (MAIN, "explain", "explain"),
        (MAIN, "explain", "explain analyze"),
        (MAIN, "explain", "explain using text"),
        (MAIN, "grant", "grant"),
        (MAIN, "grant", "revoke"),
        (MAIN, "unsupported_ddl", "create"),
        (MAIN, "unsupported_ddl", "select\ninto"),
        (MAIN, "unsupported_ddl", "insert"),
        (MAIN, "unsupported_ddl", "insert into"),
        (MAIN, "unsupported_ddl", "insert overwrite"),
        (MAIN, "unsupported_ddl", "insert overwrite into"),
        (MAIN, "unsupported_ddl", "update"),
        (MAIN, "name", "my_table_45"),
        (MAIN, "name", "replace"),
        (MAIN, "other_identifiers", "$2"),
        (MAIN, "other_identifiers", "?2"),
        (MAIN, "other_identifiers", "$email"),
        (MAIN, "other_identifiers", "METADATA$FILENAME"),
        (MAIN, "other_identifiers", "METADATA$FILE_LAST_MODIFIED"),
        (MAIN, "other_identifiers", "@my_unquoted_stage"),
        (MAIN, "other_identifiers", "%s"),
        (MAIN, "other_identifiers", "%(name)s"),
        (MAIN, "other_identifiers", "%(anything! else!)s"),
        (MAIN, "star_columns", "*columns"),
        (MAIN, "newline", "\n"),
        (JINJA, "jinja_comment", "{# my comment #}"),
        (JINJA, "jinja_comment", "{#-my comment -#}"),
        (JINJA, "jinja_comment", "{#-\nmy\ncomment\n-#}"),
        (JINJA, "jinja_statement_start", "{%"),
        (JINJA, "jinja_statement_start", "{%-"),
        (JINJA, "jinja_expression_start", "{{"),
        (JINJA, "jinja_expression_start", "{{-"),
        (JINJA, "jinja_statement_end", "%}"),
        (JINJA, "jinja_statement_end", "-%}"),
        (JINJA, "jinja_expression_end", "}}"),
        (JINJA, "jinja_expression_end", "-}}"),
        (JINJA, "jinja_set_block_start", "{% set foo %}"),
        (JINJA, "jinja_set_block_start", "{% set my_long_variable %}"),
        (JINJA, "jinja_set_block_start", "{% set ns.my_namespace_var %}"),
        (JINJA, "jinja_set_block_end", "{% endset %}"),
        (JINJA, "jinja_set_block_end", "{%- endset %}"),
        (JINJA, "jinja_if_block_start", "{% if bar %}"),
        (JINJA, "jinja_if_block_start", "{%- if bar = baz -%}"),
        (JINJA, "jinja_if_block_start", "{%- if is_incremental() -%}"),
        (JINJA, "jinja_if_block_start", "{%- if loop.last -%}"),
        (JINJA, "jinja_elif_block_start", "{%- elif 1 > 2 -%}"),
        (JINJA, "jinja_else_block_start", "{% else %}"),
        (JINJA, "jinja_for_block_start", "{%- for foo in bar %}"),
        (JINJA, "jinja_macro_block_start", "{% macro my_macro(arg1, arg2) %}"),
        (
            JINJA,
            "jinja_test_block_start",
            "{% test my_test(model, column_name) %}",
        ),
        (JINJA, "jinja_if_block_end", "{% endif %}"),
        (JINJA, "jinja_if_block_end", "{%- endif -%}"),
        (JINJA, "jinja_for_block_end", "{% endfor %}"),
        (JINJA, "jinja_macro_block_end", "{% endmacro %}"),
        (JINJA, "jinja_test_block_end", "{% endtest %}"),
        (JINJA, "jinja_snapshot_block_start", "{% snapshot snp_my_snapshot %}"),
        (JINJA, "jinja_snapshot_block_end", "{% endsnapshot %}"),
        (
            JINJA,
            "jinja_materialization_block_start",
            "{% materialization table, default %}",
        ),
        (
            JINJA,
            "jinja_materialization_block_start",
            "{% materialization incremental, adapter='redshift' %}",
        ),
        (
            JINJA,
            "jinja_materialization_block_start",
            '{% materialization incremental, adapter = "bigquery" %}',
        ),
        (JINJA, "jinja_materialization_block_end", "{% endmaterialization %}"),
        (
            JINJA,
            "jinja_call_statement_block_start",
            "{% call statement('period_boundaries', fetch_result=True) -%}",
        ),
        (JINJA, "jinja_call_statement_block_start", "{%- call statement() %}"),
        (JINJA, "jinja_call_statement_block_start", "{% call statement('main') -%}"),
        (JINJA, "jinja_call_block_start", "{% call dbt_pkg.foo('main', arg='bar') -%}"),
        (
            JINJA,
            "jinja_call_block_start",
            "{% call(table) macro_name(list_of_tables) -%}",
        ),
        (JINJA, "jinja_call_block_end", "{%- endcall %}"),
        (GRANT, "unterm_keyword", "grant"),
        (GRANT, "unterm_keyword", "revoke"),
        (GRANT, "unterm_keyword", "revoke grant option for"),
        (GRANT, "unterm_keyword", "to"),
        (GRANT, "unterm_keyword", "from"),
        (GRANT, "unterm_keyword", "with grant option"),
        (MAIN, "create_function", "CREATE OR REPLACE TABLE FUNCTION"),
        (MAIN, "create_function", "CREATE secure external function"),
        (MAIN, "create_function", "ALTER FUNCTION"),
        (MAIN, "create_function", "DROP FUNCTION IF EXISTS"),
        (FUNCTION, "unterm_keyword", "create temp function"),
        (FUNCTION, "unterm_keyword", "CREATE OR REPLACE TABLE FUNCTION"),
        (FUNCTION, "unterm_keyword", "CREATE OR REPLACE external FUNCTION"),
        (FUNCTION, "unterm_keyword", "ALTER FUNCTION"),
        (FUNCTION, "unterm_keyword", "ALTER FUNCTION IF EXISTS"),
        (FUNCTION, "unterm_keyword", "drop function"),
        (FUNCTION, "unterm_keyword", "language"),
        (FUNCTION, "unterm_keyword", "called on null input"),
        (FUNCTION, "unterm_keyword", "returns\nnull on null input"),
        (FUNCTION, "unterm_keyword", "not null"),
        (FUNCTION, "unterm_keyword", "handler"),
        (FUNCTION, "unterm_keyword", "packages"),
        (FUNCTION, "unterm_keyword", "set schema"),
        (FUNCTION, "unterm_keyword", "set secure"),
        (FUNCTION, "unterm_keyword", "set comment"),
        (FUNCTION, "unterm_keyword", "unset comment"),
        (FUNCTION, "unterm_keyword", "reset all"),
        (FUNCTION, "unterm_keyword", "no DEPENDS ON EXTENSION"),
        (FUNCTION, "unterm_keyword", "cascade"),
        (FUNCTION, "unterm_keyword", "restrict"),
        (FUNCTION, "unterm_keyword", "api_integration"),
        (FUNCTION, "unterm_keyword", "set headers"),
        (MAIN, "create_warehouse", "create warehouse if not exists"),
        (MAIN, "create_warehouse", "alter warehouse if exists"),
        (WAREHOUSE, "unterm_keyword", "create warehouse if not exists"),
        (WAREHOUSE, "unterm_keyword", "create or replace warehouse"),
        (WAREHOUSE, "unterm_keyword", "warehouse_type"),
        (WAREHOUSE, "unterm_keyword", "with warehouse_size"),
        (WAREHOUSE, "unterm_keyword", "set warehouse_size"),
        (WAREHOUSE, "unterm_keyword", "max_cluster_count"),
        (WAREHOUSE, "unterm_keyword", "min_cluster_count"),
        (WAREHOUSE, "unterm_keyword", "auto_suspend"),
        (WAREHOUSE, "unterm_keyword", "auto_resume"),
        (WAREHOUSE, "unterm_keyword", "alter warehouse if exists"),
        (WAREHOUSE, "unterm_keyword", "rename to"),
        (WAREHOUSE, "unterm_keyword", "set tag"),
        (WAREHOUSE, "unterm_keyword", "resume if suspended"),
        (WAREHOUSE, "unterm_keyword", "unset scaling_policy"),
        (MAIN, "create_clone", "create table foo clone"),
        (MAIN, "create_clone", "create table db.sch.foo clone"),
        (MAIN, "create_clone", "create or replace database foo clone"),
        (MAIN, "create_clone", "create stage if not exists foo clone"),
        (CLONE, "unterm_keyword", "create table"),
        (CLONE, "unterm_keyword", "create database"),
        (CLONE, "unterm_keyword", "create schema"),
        (CLONE, "unterm_keyword", "create\nfile\nformat"),
        (CLONE, "unterm_keyword", "clone"),
        (CLONE, "name", "foo"),
        (CLONE, "word_operator", "at"),
        (CLONE, "word_operator", "before"),
        (
            UNSUPPORTED,
            "unsupported_line",
            "create table foo (a text default 'foo;bar')",
        ),
    ],
)
def test_regex_exact_match(
    ruleset: List[Rule],
    rule_name: str,
    value: str,
) -> None:
    rule = get_rule(ruleset, rule_name)
    match = rule.program.match(value)
    assert match is not None, f"{rule_name} regex doesn't match {value}"
    start, end = match.span(1)

    assert value[start:end] == value, f"{rule_name} regex doesn't exactly match {value}"


@pytest.mark.parametrize(
    "ruleset,rule_name,value",
    [
        (CORE, "fmt_off", "# fmt:"),
        (CORE, "fmt_off", "-- fmt: off but not really"),
        (CORE, "jinja_start", "{"),
        (CORE, "jinja_end", "}"),
        (CORE, "quoted_name", "my_unquoted_name"),
        (CORE, "quoted_name", "$bar$foo$baz$"),
        (CORE, "double_colon", ":"),
        (CORE, "other_identifiers", "?"),
        (CORE, "operator", "."),
        (MAIN, "word_operator", "using"),
        (MAIN, "word_operator", "any"),
        (MAIN, "word_operator", "all"),
        (MAIN, "unterm_keyword", "lateral flatten"),
        (MAIN, "unterm_keyword", "for"),
        (MAIN, "unterm_keyword", "MAIN into"),
        (MAIN, "star_replace_exclude", "replace"),
        (MAIN, "unterm_keyword", "MAINion"),
        (MAIN, "unterm_keyword", "delete"),
        (MAIN, "unsupported_ddl", "insert('abc', 1, 2, 'Z')"),
        (MAIN, "unsupported_ddl", "get(foo, 'bar')"),
        (MAIN, "create_clone", "create table"),
        (JINJA, "jinja_set_block_start", "{% set foo = 'baz' %}"),
        (JINJA, "jinja_call_statement_block_start", "{% call(t) statement('main') -%}"),
        (GRANT, "unterm_keyword", "select"),
        (FUNCTION, "unterm_keyword", "secure"),
    ],
)
def test_regex_anti_match(
    ruleset: List[Rule],
    rule_name: str,
    value: str,
) -> None:
    """make sure our compiled programs do not match these values"""
    rule = get_rule(ruleset, rule_name)
    match = rule.program.match(value)
    assert match is None, f"{rule_name} regex should not match {value}"


@pytest.mark.parametrize(
    "ruleset,rule_name,value,matched_value",
    [
        (MAIN, "frame_clause", "rows between unbounded preceding", "rows "),
        (MAIN, "frame_clause", "rows unbounded preceding", "rows "),
        (MAIN, "frame_clause", "rows 1 preceding", "rows "),
        (MAIN, "frame_clause", "range between current row", "range "),
        (MAIN, "frame_clause", "range 1 following", "range "),
        (MAIN, "frame_clause", "range current row", "range "),
        (MAIN, "frame_clause", "groups between 1 preceding", "groups "),
        (
            MAIN,
            "functions_that_overlap_with_word_operators",
            "filter(foo)",
            "filter",
        ),
        (
            MAIN,
            "functions_that_overlap_with_word_operators",
            "isnull(bar, baz)",
            "isnull",
        ),
        (
            MAIN,
            "functions_that_overlap_with_word_operators",
            "like('foo', 'bar')",
            "like",
        ),
        (
            MAIN,
            "functions_that_overlap_with_word_operators",
            "rlike('foo', 'bar')",
            "rlike",
        ),
        (
            MAIN,
            "functions_that_overlap_with_word_operators",
            "ilike('foo', 'bar')",
            "ilike",
        ),
        (
            UNSUPPORTED,
            "unsupported_line",
            "create table foo (a -- my comment",
            "create table foo (a",
        ),
        (
            UNSUPPORTED,
            "unsupported_line",
            "create table foo (a {{ some_jinja }}",
            "create table foo (a",
        ),
    ],
)
def test_regex_partial_match(
    ruleset: List[Rule], rule_name: str, value: str, matched_value: str
) -> None:
    rule = get_rule(ruleset, rule_name)
    match = rule.program.match(value)
    assert match is not None, f"{rule_name} regex doesn't match {value}"
    start, end = match.span(1)

    assert value[start:end] == matched_value, (
        f"{rule_name} regex doesn't exactly match {matched_value}"
    )


def test_regex_should_not_match_empty_string() -> None:
    rules = itertools.chain.from_iterable(ALL_RULESETS)
    for rule in rules:
        match = rule.program.match("")
        assert match is None, f"{rule.name} rule matches empty string"


def test_core_priority_range() -> None:
    for rule in CORE:
        assert rule.priority >= 0
        assert rule.priority < 10000
        #  this range reserved for other rulesets
        assert not (rule.priority > 1000 and rule.priority < 5000)


@pytest.mark.parametrize("ruleset", ALL_RULESETS)
def test_rule_names_unique_within_ruleset(ruleset: List[Rule]) -> None:
    name_counts = Counter([rule.name for rule in ruleset])
    assert max(name_counts.values()) == 1


@pytest.mark.parametrize("ruleset", ALL_RULESETS)
def test_rule_priorities_unique_within_ruleset(ruleset: List[Rule]) -> None:
    priority_counts = Counter([rule.priority for rule in ruleset])
    assert max(priority_counts.values()) == 1


# unsupported ruleset intentionally includes a dupe rule for names
@pytest.mark.parametrize("ruleset", ALL_RULESETS[:-1])
def test_rule_patterns_unique_within_ruleset(ruleset: List[Rule]) -> None:
    pattern_counts = Counter([rule.pattern for rule in ruleset])
    assert max(pattern_counts.values()) == 1



import re
from typing import List, Optional, Tuple

from sqlfmt.exception import SqlfmtBracketError
from sqlfmt.line import Line
from sqlfmt.node import Node, get_previous_token
from sqlfmt.tokens import Token, TokenType


class NodeManager:
    def __init__(self, case_sensitive_names: bool) -> None:
        self.case_sensitive_names = case_sensitive_names

    def create_node(self, token: Token, previous_node: Optional[Node]) -> Node:
        """
        Create a Node from a Token. For the Node to be properly formatted,
        method call must also include a reference to the previous Node in the
        Query (either on the same or previous Line), unless it is the first
        Node in the Query.

        The Node's depth and whitespace are calculated when it is created
        (this does most of the formatting of the Node). Node values are
        lowercased if they are simple names, keywords, or statements.
        """

        open_brackets, open_jinja_blocks = self.open_brackets(token, previous_node)
        formatting_disabled = self.disable_formatting(token, previous_node)
        if formatting_disabled:
            prefix = token.prefix
            value = token.token
        else:
            prev_token, extra_whitespace = get_previous_token(previous_node)
            prefix = self.whitespace(token, prev_token, extra_whitespace)
            value = self.standardize_value(token)

        return Node(
            token=token,
            previous_node=previous_node,
            prefix=prefix,
            value=value,
            open_brackets=open_brackets,
            open_jinja_blocks=open_jinja_blocks,
            formatting_disabled=formatting_disabled,
        )

    def raise_on_mismatched_bracket(self, token: Token, last_bracket: Node) -> None:
        """
        Raise a SqlfmtBracketError if token is a closing bracket, but it
        does not match the token in the last_bracket node
        """
        matches = {
            "{": "}",
            "(": ")",
            "[": "]",
            "case": "end",
            "array<": ">",
            "map<": ">",
            "table<": ">",
            "struct<": ">",
        }
        last_bracket_value = last_bracket.value.lower()
        if (
            last_bracket.token.type
            not in (TokenType.BRACKET_OPEN, TokenType.STATEMENT_START)
            or last_bracket_value not in matches
            or matches[last_bracket_value] != token.token.lower()
        ):
            raise SqlfmtBracketError(
                f"Closing bracket '{token.token}' found at {token.spos} does not "
                f"match last opened bracket '{last_bracket.value}' found at "
                f"{last_bracket.token.spos}."
            )

    def raise_on_mismatched_jinja_tags(self, token: Token, start_tag: Node) -> None:
        """
        Compare the value of token to the start_tag to determine whether token
        closes start_tag
        """
        try:
            if "endif" in token.token.lower():
                if not any(s in start_tag.value for s in ["if", "elif", "else"]):
                    raise ValueError
            elif "endfor" in token.token.lower():
                if not any(s in start_tag.value for s in ["for", "else"]):
                    raise ValueError
            else:
                end_text, _ = re.subn(r"[{}%\-\s]", "", token.token.lower())
                start_value = end_text.replace("end", "")
                if start_value not in start_tag.value:
                    raise ValueError
        except ValueError as e:
            raise SqlfmtBracketError(
                f"Closing jinja tag '{token.token}' found at pos {token.spos} does "
                f"not match last opened tag '{start_tag.value}' found at pos "
                f"{start_tag.token.spos}."
            ) from e

    def open_brackets(
        self, token: Token, previous_node: Optional[Node]
    ) -> Tuple[List[Node], List[Node]]:
        """
        Uses the previous_node and the contents of the current token
        to compute the depth of the new node.

        Returns two lists, for open_brackets and open_jinja_blocks
        """

        if previous_node is None:
            open_brackets = []
            open_jinja_blocks = []
        else:
            open_brackets = previous_node.open_brackets.copy()
            open_jinja_blocks = previous_node.open_jinja_blocks.copy()

            # add the previous node to the list of open brackets or jinja blocks
            if previous_node.is_unterm_keyword or previous_node.is_opening_bracket:
                open_brackets.append(previous_node)
            elif previous_node.is_opening_jinja_block:
                open_jinja_blocks.append(previous_node)

        # if the token should reduce the depth of the node, pop
        # the last item(s) off open_brackets or open_jinja_blocks
        if token.type in (TokenType.UNTERM_KEYWORD, TokenType.SET_OPERATOR):
            if open_brackets and open_brackets[-1].is_unterm_keyword:
                _ = open_brackets.pop()
        elif token.type in (TokenType.BRACKET_CLOSE, TokenType.STATEMENT_END):
            try:
                last_bracket = open_brackets.pop()
                if last_bracket.is_unterm_keyword:
                    last_bracket = open_brackets.pop()
            except IndexError as e:
                raise SqlfmtBracketError(
                    f"Closing bracket '{token.token}' found at "
                    f"{token.spos} before bracket was opened."
                ) from e
            else:
                self.raise_on_mismatched_bracket(token, last_bracket)
        elif token.type is TokenType.JINJA_BLOCK_END:
            try:
                start_tag = open_jinja_blocks.pop()
                self.raise_on_mismatched_jinja_tags(token, start_tag)
            except IndexError as e:
                raise SqlfmtBracketError(
                    f"Closing bracket '{token.token}' found at "
                    f"{token.spos} before bracket was opened."
                ) from e
        # if we hit a semicolon, reset open_brackets, since we're
        # about to start a new query
        elif token.type is TokenType.SEMICOLON:
            open_brackets = []

        return open_brackets, open_jinja_blocks

    def whitespace(
        self,
        token: Token,
        previous_token: Optional[Token],
        extra_whitespace: bool,
    ) -> str:
        """
        Returns the proper whitespace before the token literal, to be set as the
        prefix of the Node.

        Most tokens should be prefixed by a simple space. Other cases are outlined
        below.
        """
        NO_SPACE = ""
        SPACE = " "

        # tokens that are never preceded by a space
        if token.type.is_never_preceded_by_space:
            return NO_SPACE
        # no spaces after an open bracket or a cast operator (::)
        elif previous_token and previous_token.type in (
            TokenType.BRACKET_OPEN,
            TokenType.DOUBLE_COLON,
        ):
            return NO_SPACE
        # always a space before a keyword
        elif token.type.is_preceded_by_space_except_after_open_bracket:
            return SPACE
        # names preceded by dots or colons are namespaced identifiers. No space.
        elif (
            token.type.is_possible_name
            and previous_token
            and previous_token.type in (TokenType.DOT, TokenType.COLON)
        ):
            return NO_SPACE
        # numbers preceded by colons are simple slices. No Space
        elif (
            token.type is TokenType.NUMBER
            and previous_token
            and previous_token.type is TokenType.COLON
        ):
            return NO_SPACE
        # open brackets that contain `<` are bq type definitions
        # like `array<` in `array<int64>` and require a space,
        # unless the preceding token is also an open bracket
        elif token.type is TokenType.BRACKET_OPEN and "<" in token.token:
            if previous_token and previous_token.type is not TokenType.BRACKET_OPEN:
                return SPACE
            else:
                return NO_SPACE
        # open brackets that follow names are function calls or array indexes.
        # open brackets that follow closing brackets are array indexes.
        # open brackets that follow open brackets are just nested brackets.
        # No Space.
        elif (
            token.type is TokenType.BRACKET_OPEN
            and previous_token
            and previous_token.type
            in (
                TokenType.NAME,
                TokenType.QUOTED_NAME,
                TokenType.BRACKET_OPEN,
                TokenType.BRACKET_CLOSE,
            )
        ):
            return NO_SPACE
        # open square brackets that follow colons are escaped databricks
        # variant cols
        elif (
            token.type is TokenType.BRACKET_OPEN
            and token.token == "["
            and previous_token
            and previous_token.type is TokenType.COLON
        ):
            return NO_SPACE
        # need a space before any other open bracket
        elif token.type is TokenType.BRACKET_OPEN:
            return SPACE
        # we don't know what a jinja expression will evaluate to,
        # so we have to respect the original text
        elif token.type.is_jinja:
            if token.prefix != "" or extra_whitespace:
                return SPACE
            else:
                return NO_SPACE
        elif previous_token and previous_token.type is TokenType.JINJA_EXPRESSION:
            if token.prefix != "" or extra_whitespace:
                return SPACE
            else:
                return NO_SPACE
        else:
            return SPACE

    def standardize_value(self, token: Token) -> str:
        """
        Tokens that are words (not symbols) and aren't jinja
        or comments should be lowercased and have any internal
        whitespace replaced with a single space
        """
        if token.type.is_always_lowercased:
            return " ".join(token.token.lower().split())
        elif token.type is TokenType.NAME and not self.case_sensitive_names:
            return token.token.lower()
        else:
            return token.token

    def disable_formatting(
        self, token: Token, previous_node: Optional[Node]
    ) -> List[Token]:
        """
        Manage the formatting_disabled property for the node to be created from
        the token and previous node.
        """
        formatting_disabled = (
            previous_node.formatting_disabled.copy()
            if previous_node is not None
            else []
        )

        if token.type in (TokenType.FMT_OFF, TokenType.DATA):
            formatting_disabled.append(token)

        if (
            formatting_disabled
            and previous_node is not None
            and previous_node.token.type
            in (
                TokenType.FMT_ON,
                TokenType.DATA,
            )
        ):
            formatting_disabled.pop()

        return formatting_disabled

    def append_newline(self, line: Line) -> None:
        """
        Create a new NEWLINE token and append it to the end of line
        """
        previous_node: Optional[Node] = None
        previous_token: Optional[Token] = None
        if line.nodes:
            previous_node = line.nodes[-1]
            previous_token = line.nodes[-1].token
        elif line.previous_node is not None:
            previous_node = line.previous_node
            previous_token = line.previous_node.token

        if previous_token:
            spos = previous_token.epos
            epos = spos
        else:
            spos = 0
            epos = 0

        nl = Token(
            type=TokenType.NEWLINE,
            prefix="",
            token="\n",
            spos=spos,
            epos=epos,
        )

        node = self.create_node(token=nl, previous_node=previous_node)
        line.nodes.append(node)



import re
from dataclasses import dataclass
from typing import ClassVar, Iterator, Optional, Tuple

from sqlfmt.node import Node
from sqlfmt.tokens import Token, TokenType


@dataclass
class Comment:
    """
    A Comment wraps a token (of type COMMENT), and provides a number of properties and
    methods that are used in formatting and printing the query
    """

    token: Token
    is_standalone: bool
    previous_node: Optional[Node]
    comment_marker: ClassVar[re.Pattern] = re.compile(r"(--|#|//|/\*|\{#-?)([^\S\n]*)")

    def __str__(self) -> str:
        """
        Returns the contents of the comment token,
        without preceding whitespace, with a single space between the marker
        and the comment text.
        """
        if (
            self.is_multiline
            or self.formatting_disabled
            or self.is_databricks_query_hint
        ):
            return self.token.token
        else:
            marker, comment_text = self._comment_parts()
            if comment_text:
                return f"{marker} {comment_text}"
            else:
                return marker

    def __len__(self) -> int:
        return len(str(self))

    def _get_marker(self) -> Tuple[str, int]:
        """
        For a comment, returns a tuple.

        The first element is the comment's marker, which is the symbol or symbols
        that indicates that the rest of the token is a comment; e.g., "--" or "#"

        The second element is the position of the comment's text, which is the
        first non-whitespace character after the marker
        """
        match = self.comment_marker.match(self.token.token)
        assert match, f"{self.token.token} does not match comment marker"
        _, epos = match.span(1)
        _, text_offset = match.span(2)
        return self.token.token[:epos], text_offset

    def _rewrite_marker(self, marker: str) -> str:
        """
        Rewrites the comment marker to the standard --

        The following markers are rewritten:
            //
        """
        if marker == "//":
            return "--"
        else:
            return marker

    def _comment_parts(self) -> Tuple[str, str]:
        """
        For a comment, returns a tuple of the comment's marker and its contents
        (without leading whitespace)
        """
        assert not self.is_multiline
        marker, skipchars = self._get_marker()
        comment_text = self.token.token[skipchars:]
        return self._rewrite_marker(marker), comment_text

    @property
    def is_multiline(self) -> bool:
        """
        Returns True if this Comment contains newlines
        """
        return "\n" in self.token.token

    @property
    def is_c_style(self) -> bool:
        return self.token.token.startswith("/*")

    @property
    def is_databricks_query_hint(self) -> bool:
        return self.token.token.startswith("/*+")

    @property
    def is_inline(self) -> bool:
        return not self.is_standalone and not self.is_multiline and not self.is_c_style

    @property
    def body(self) -> str:
        if self.is_multiline:
            return self.token.token.strip()
        else:
            return self._comment_parts()[1].strip()

    @property
    def formatting_disabled(self) -> bool:
        if self.previous_node is None:
            return False
        else:
            # comment formatting is only disabled if there is an explicit FMT_OFF token
            # (i.e., not if node formatting is disabled due to DATA nodes).
            return any(
                [
                    t.type is TokenType.FMT_OFF
                    for t in self.previous_node.formatting_disabled
                ]
            )

    @property
    def should_preserve_indentation(self) -> bool:
        """
        Comments should preserve their original (token) indentation
        if they have formatting disabled, OR if they are in a run
        of DATA tokens.

        Newlines in runs of DATA tokens won't themselves have
        formatting disabled, so it's insufficient to check
        the previous token.
        """
        return self.formatting_disabled or (
            self.previous_node is not None
            and self.previous_node.is_newline
            and self.previous_node.previous_node is not None
            and self.previous_node.previous_node.token.type is TokenType.DATA
        )

    def render_inline(self) -> str:
        """
        Renders a comment as an inline comment, assuming it'll fit.
        """
        prefix = self.token.prefix if self.formatting_disabled else "  "
        return f"{prefix}{self}"

    def render_standalone(self, max_length: int, prefix: str) -> str:
        """
        For a Comment, returns the string for properly formatting this Comment
        as a standalone comment (on its own line)
        """
        if self.should_preserve_indentation:
            prefix = self.token.prefix

        if self.formatting_disabled or self.is_multiline:
            rendered = f"{prefix}{self}"
        else:
            # split long standalone comments onto multiple lines
            if len(self) + len(prefix) <= max_length:
                rendered = prefix + str(self)
            else:
                marker, comment_text = self._comment_parts()
                if marker in ("--", "#"):
                    available_length = max_length - len(prefix) - len(marker) - 2
                    line_gen = self._split_before(comment_text, available_length)
                    rendered = "".join(
                        [prefix + marker + " " + txt.strip() + "\n" for txt in line_gen]
                    )
                else:  # block-style or jinja comment. Don't wrap long lines for now
                    rendered = prefix + str(self)
        nl = "\n"
        return f"{rendered.rstrip(nl)}\n"

    @classmethod
    def _split_before(cls, text: str, max_length: int) -> Iterator[str]:
        """
        When rendering very long comments, we try to split them at the desired line
        length and wrap them onto multiple lines. This method takes the contents of
        a comment (without the marker) and a maximum length, and splits the original
        text at whitespace, yielding each split as a stringd
        """
        if len(text) < max_length:
            yield text.rstrip()
        elif re.match(r".*({{.*?}}|{%.*?%}).*", text):
            # jinja comments are not split
            yield text.rstrip()
        else:
            for idx, char in enumerate(reversed(text[:max_length])):
                if char.isspace():
                    yield text[: max_length - idx].rstrip()
                    yield from cls._split_before(text[max_length - idx :], max_length)
                    break
            else:  # no spaces in the comment
                yield text.rstrip()



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import group
from sqlfmt.rules.core import CORE
from sqlfmt.tokens import TokenType

GRANT = [
    *CORE,
    Rule(
        name="unterm_keyword",
        priority=1300,
        pattern=group(
            r"grant",
            r"revoke(\s+grant\s+option\s+for)?",
            r"on",
            r"to",
            r"from",
            r"with\s+grant\s+option",
            r"granted\s+by",
            r"cascade",
            r"restrict",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
]



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import PRAGMA_SET_CALL, group
from sqlfmt.rules.core import CORE
from sqlfmt.tokens import TokenType

# Covers simple PRAGMA, SET, and CALL statements, e.g.:
# PRAGMA database_list;
# PRAGMA storage_info('table_name');
# CALL pragma_storage_info('table_name');
# SET default_collation = 'nocase';
PRAGMA = [
    *CORE,
    Rule(
        name="unterm_keyword",
        priority=1300,
        pattern=group(PRAGMA_SET_CALL) + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
]



import pytest

from sqlfmt.analyzer import Analyzer


def test_whitespace_formatting(default_analyzer: Analyzer) -> None:
    source_string = "  select 1\n    from my_table\nwhere true"
    expected_string = "select 1\nfrom my_table\nwhere true\n"
    q = default_analyzer.parse_query(source_string=source_string)
    assert str(q) == expected_string


def test_only_comment_formatting(default_analyzer: Analyzer) -> None:
    source_string = "-- a comment"
    expected_string = "-- a comment\n"
    q = default_analyzer.parse_query(source_string=source_string)
    assert str(q) == expected_string


@pytest.mark.parametrize("source_string", ["", "\n"])
def test_empty_formatting(default_analyzer: Analyzer, source_string: str) -> None:
    expected_string = ""
    q = default_analyzer.parse_query(source_string=source_string)
    assert str(q) == expected_string



def group(*choices: str) -> str:
    """
    Convenience function for creating grouped alternatives in regex
    """
    return f"({'|'.join(choices)})"


NEWLINE: str = r"\r?\n"
EOL = group(NEWLINE, r"$")

JINJA_START = group(r"\{[{%#]")

SQL_QUOTED_EXP = group(
    # tripled single quotes (optionally raw/bytes).
    #
    # The body cannot be a lazy .*? — that closes on the FIRST ''' it can reach,
    # so `select '''' || 'x' || ''''` is read as a triple-quoted string starting
    # at the first quote and ending three quotes later, stranding a lone quote
    # that nothing can lex. Matching quote runs explicitly, and refusing a
    # closer that is itself followed by another quote, makes the alternative
    # decline rather than mis-close, so the escaped-single-quote alternative
    # below gets its turn. [^'] rather than [^'\r\n] because these patterns are
    # compiled with re.DOTALL and multi-line triple-quoted strings must keep
    # matching.
    r"(rb?|b|br)?'''(?:[^'](?:[^']|'(?!''))*)?'''(?!')",
    # tripled double quotes follow the same quote-run rule as single quotes.
    r'(rb?|b|br)?"""(?:[^"](?:[^"]|"(?!""))*)?"""(?!")',
    # possibly escaped double quotes
    r'(rb?|b|br|u&|@)?"([^"\\]*(\\.[^"\\]*|""[^"\\]*)*)"',
    # possibly escaped single quotes
    r"(rb?|b|br|u&|x)?'([^'\\]*(\\.[^'\\]*|''[^'\\]*)*)'",
    r"\$(?P<tag>\w*)\$.*?\$(?P=tag)\$",  # pg dollar-delimited strings
    # possibly escaped backtick
    r"`([^`\\]*(\\.[^`\\]*)*)`",
)

SQL_COMMENT_START = r"(?=--|#|//|/\*)"
SQL_COMMENT = group(
    r"--[^\r\n]*",
    r"#[^\r\n]*",
    r"//[^\r\n]*",  # snowflake's js-style double-slash comment
    r"/\*[^*]*\*+(?:[^/*][^*]*\*+)*/",  # simple block comment
)

CREATE_FUNCTION = (
    r"create(\s+or\s+replace)?(\s+temp(orary)?)?(\s+secure)?"
    r"(\s+external)?(\s+table)?"
    r"\s+function(\s+if\s+not\s+exists)?"
)
ALTER_DROP_FUNCTION = r"(alter|drop)\s+function(\s+if\s+exists)?"

CREATE_WAREHOUSE = r"create(\s+or\s+replace)?\s+warehouse(\s+if\s+not\s+exists)?"
ALTER_WAREHOUSE = r"alter\s+warehouse(\s+if\s+exists)?"

CREATE_CLONABLE = (
    r"create(\s+or\s+replace)?\s+"
    + group(
        r"database",
        r"schema",
        r"table",
        r"stage",
        r"file\s+format",
        r"sequence",
        r"stream",
        r"task",
    )
    + r"(\s+if\s+not\s+exists)?"
)

PRAGMA_SET_CALL = group(r"pragma", r"set", r"call")



from typing import List

import pytest

from sqlfmt.analyzer import Analyzer
from sqlfmt.exception import SqlfmtSegmentError
from sqlfmt.line import Line
from sqlfmt.segment import Segment, create_segments_from_lines
from tests.util import read_test_data


@pytest.mark.parametrize(
    "source_string,expected_result",
    [
        ("case\nfoo\nend\n", True),
        ("count(\n*\n)\n", True),
        ("count(\n*\n)\n\n\n", True),
        ("\n\n\ncount(\n*\n)", True),
        ("as (\nselect\n1\n)\n", True),
        ("as foo\n", False),
        ("from\nfoo\nwhere\n", False),
    ],
)
def test_tail_closes_head(
    default_analyzer: Analyzer, source_string: str, expected_result: bool
) -> None:
    q = default_analyzer.parse_query(source_string)
    segment = Segment(q.lines)
    assert segment.tail_closes_head == expected_result


def test_tail_closes_head_empty(default_analyzer: Analyzer) -> None:
    q = default_analyzer.parse_query("a\n\n\n\n\n\n\n\nb\n")
    blank_lines = q.lines[1:-1]
    assert all(line.is_blank_line for line in blank_lines)
    segment = Segment(blank_lines)
    assert segment.tail_closes_head is False


@pytest.mark.parametrize(
    "source_string,expected_idx",
    [
        ("case\nfoo\nend\n", 0),
        ("\n\n\ncount(\n*\n)", 3),
        ("\n\n\n    count(\n*\n)", 3),
        ("\n     \n\n    count(\n*\n)", 3),
    ],
)
def test_segment_head(
    default_analyzer: Analyzer, source_string: str, expected_idx: int
) -> None:
    q = default_analyzer.parse_query(source_string)
    segment = Segment(q.lines)
    line, i = segment.head
    assert i == expected_idx
    assert line == q.lines[i]


@pytest.mark.parametrize(
    "source_string",
    [
        "",
        "\n",
        "\n\n\n    \n\n",
    ],
)
def test_segment_head_raises(default_analyzer: Analyzer, source_string: str) -> None:
    q = default_analyzer.parse_query(source_string)
    segment = Segment(q.lines)
    with pytest.raises(SqlfmtSegmentError):
        _, _ = segment.head


@pytest.mark.parametrize(
    "source_string,expected_idx",
    [
        ("case\nfoo\nend,\na\n", 0),
        ("count(\n*\n),\n\n\na\n", 2),
    ],
)
def test_segment_tail(
    default_analyzer: Analyzer, source_string: str, expected_idx: int
) -> None:
    q = default_analyzer.parse_query(source_string)
    # the parser eliminates trailing empty lines, so
    # we need to add a nonblank line to our test cases
    # and then strip that away
    segment = Segment(q.lines[:-1])
    line, i = segment.tail
    assert i == expected_idx
    assert line == segment[-(i + 1)]


@pytest.mark.parametrize(
    "source_string",
    [
        "",
        "\n",
        "\n\n\n    \n\n",
    ],
)
def test_segment_tail_raises(default_analyzer: Analyzer, source_string: str) -> None:
    q = default_analyzer.parse_query(source_string)
    segment = Segment(q.lines)
    with pytest.raises(SqlfmtSegmentError):
        _, _ = segment.tail


def test_split_into_segments(default_analyzer: Analyzer) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_merger/test_split_into_segments.sql"
    )
    q = default_analyzer.parse_query(source_string)

    top_level_segments = create_segments_from_lines(q.lines)

    assert len(top_level_segments) == 2
    assert str(top_level_segments[0][0]).startswith("select")
    assert str(top_level_segments[1][0]).startswith("from")

    select_segment = [top_level_segments[0]]

    assert select_segment == create_segments_from_lines(select_segment[0])

    select_lines = select_segment[0]
    assert not (
        select_lines[-1].closes_bracket_from_previous_line
        and select_lines[-1].depth == select_lines[0].depth
    )

    indented_segments = create_segments_from_lines(select_lines[1:])
    expected_segments = [
        "    my_first_field,\n",
        "    my_second_field\n",
        "    as an_alias,\n",
        "    case\n",
        "    as my_case_statement,\n",
        "    case\n",
        "    case\n",
        "    ::numeric(\n",
        "    as casted_case\n",
        "    ,\n",
        "    (\n",
        "    +\n",
        "    ::varchar(\n",
        "    another_field,\n",
        "    case\n",
        "    + 4\n",
    ]
    assert [str(segment[0]) for segment in indented_segments] == expected_segments


def test_split_into_segments_empty() -> None:
    no_lines: List[Line] = []
    result = create_segments_from_lines(no_lines)
    assert result == []


@pytest.mark.parametrize(
    "source_string,idx,expected_len",
    [
        ("count(\nfoo\n)\n", 0, 2),
        ("\n\n\n\ncount(\nfoo\n)\n", 4, 2),
        ("foo,\nbar,\nbaz\n", 0, 1),
    ],
)
def test_split_after(
    default_analyzer: Analyzer, source_string: str, idx: int, expected_len: int
) -> None:
    q = default_analyzer.parse_query(source_string)
    segment = Segment(q.lines)
    remainder = segment.split_after(idx)
    first_word = source_string.lstrip().splitlines(keepends=True)[0]
    assert first_word not in [str(line) for s in remainder for line in s]
    assert len(remainder) == expected_len



from collections import Counter
from typing import Any

import pytest

from sqlfmt.dialect import ClickHouse, Dialect, Polyglot
from sqlfmt.rules.common import group


def test_group() -> None:
    regex = group(r"a", r"b", r"\d")
    assert regex == r"(a|b|\d)"

    group_of_one = group(r"\w+")
    assert group_of_one == r"(\w+)"


def test_dialect() -> None:
    # can't instantiate abc
    with pytest.raises(TypeError):
        _ = Dialect()  # type: ignore


class TestAllDialects:
    @pytest.fixture(params=[Polyglot, ClickHouse])
    def dialect(self, request: Any) -> Polyglot:
        d = request.param()
        assert isinstance(d, Polyglot)
        return d

    def test_rule_props_are_unique(self, dialect: Polyglot) -> None:
        ruleset = dialect.get_rules()
        name_counts = Counter([rule.name for rule in ruleset])
        assert max(name_counts.values()) == 1
        priority_counts = Counter([rule.priority for rule in ruleset])
        assert max(priority_counts.values()) == 1
        pattern_counts = Counter([rule.pattern for rule in ruleset])
        assert max(pattern_counts.values()) == 1


class TestPolyglot:
    @pytest.fixture
    def polyglot(self) -> Polyglot:
        return Polyglot()

    def test_case_insensitive(self, polyglot: Polyglot) -> None:
        assert polyglot.case_sensitive_names is False


class TestClickHouse:
    @pytest.fixture
    def clickhouse(self) -> ClickHouse:
        return ClickHouse()

    def test_case_sensitive(self, clickhouse: ClickHouse) -> None:
        assert clickhouse.case_sensitive_names is True



import pytest

from sqlfmt.mode import Mode
from sqlfmt.node_manager import NodeManager
from sqlfmt.tokens import Token, TokenType
from tests.util import read_test_data


@pytest.fixture
def node_manager(default_mode: Mode) -> NodeManager:
    return NodeManager(default_mode.dialect.case_sensitive_names)


@pytest.mark.parametrize(
    "token,result", [("between", True), ("BETWEEN", True), ("like", False)]
)
def test_is_the_between_operator(
    token: str, result: bool, node_manager: NodeManager
) -> None:
    t = Token(
        type=TokenType.WORD_OPERATOR,
        prefix="",
        token=token,
        spos=0,
        epos=7,
    )
    n = node_manager.create_node(t, previous_node=None)
    assert n.is_the_between_operator is result


@pytest.mark.parametrize(
    "source_string,node_index,result",
    [
        ("my_array[0]", 1, True),
        ("my_dict['my_key']", 1, True),
        ("my_dict['my_key']['my_nested_key']", 4, True),
        ("foo, bar[0]", 3, True),
        ("foo, [0]", 2, False),
        ("foo, ['a', 'b', 'c']", 2, False),
        ("foo, [1 + foo]", 2, False),
        ('"my_quoted_array"[1]', 1, True),
        ("my_array_func()[0]", 3, True),
        ("[", 0, False),
        ("[1, 2]", 0, False),
        ("[1, 2][0]", 5, True),
    ],
)
def test_is_square_bracket_operator(
    source_string: str, node_index: int, result: bool, default_mode: Mode
) -> None:
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    n = q.nodes[node_index]
    assert n.value == "["  # just to make sure our index is right
    assert n.is_bracket_operator is result


@pytest.mark.parametrize(
    "source_string,node_index,result",
    [
        ("foo(0)", 1, False),
        ("struct<a int64, b int64>(1, 2)", 7, True),
    ],
)
def test_is_paren_bracket_operator(
    source_string: str, node_index: int, result: bool, default_mode: Mode
) -> None:
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    n = q.nodes[node_index]
    assert n.value == "("  # just to make sure our index is right
    assert n.is_bracket_operator is result


def test_is_the_and_after_the_between_operator(default_mode: Mode) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_node/test_is_the_and_after_the_between_operator.sql"
    )
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)

    and_nodes = [
        node for node in q.nodes if node.token.type is TokenType.BOOLEAN_OPERATOR
    ]
    other_nodes = [
        node for node in q.nodes if node.token.type is not TokenType.BOOLEAN_OPERATOR
    ]
    boolean_ands = and_nodes[::2]
    between_ands = and_nodes[1::2]
    assert all([not n.is_the_and_after_the_between_operator for n in boolean_ands])
    assert all([n.is_the_and_after_the_between_operator for n in between_ands])
    assert all([not n.is_the_and_after_the_between_operator for n in other_nodes])



import re
from typing import TYPE_CHECKING, Callable, List, Optional

from jinja2 import Environment

from sqlfmt.comment import Comment
from sqlfmt.exception import SqlfmtBracketError, StopRulesetLexing
from sqlfmt.line import Line
from sqlfmt.node import Node, get_previous_token
from sqlfmt.rule import MAYBE_WHITESPACES, Rule
from sqlfmt.tokens import Token, TokenType

if TYPE_CHECKING:
    from sqlfmt.analyzer import Analyzer


def group(*choices: str) -> str:
    """
    Convenience function for creating grouped alternatives in regex
    """
    return f"({'|'.join(choices)})"


def raise_sqlfmt_bracket_error(
    _: "Analyzer", source_string: str, match: re.Match
) -> None:
    spos, epos = match.span(1)
    raw_token = source_string[spos:epos]
    raise SqlfmtBracketError(
        f"Encountered closing bracket '{raw_token}' at position"
        f" {spos}, before matching opening bracket. Context:"
        f" {source_string[spos : spos + 50]}"
    )


def add_node_to_buffer(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    token_type: TokenType,
    previous_node: Optional[Node] = None,
    override_analyzer_prev_node: bool = False,
) -> None:
    """
    Create a token of token_type from the match, then create a Node
    from that token and append it to the Analyzer's buffer
    """
    if previous_node is None and override_analyzer_prev_node is False:
        previous_node = analyzer.previous_node
    token = Token.from_match(source_string, match, token_type)
    node = analyzer.node_manager.create_node(token=token, previous_node=previous_node)
    analyzer.node_buffer.append(node)
    analyzer.pos = token.epos


def safe_add_node_to_buffer(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    token_type: TokenType,
    fallback_token_type: TokenType,
) -> None:
    """
    Try to create a token of token_type from the match; if that fails
    with a SqlfmtBracketError, create a token of fallback_token_type.
    Then create a Node from that token and append it to the Analyzer's buffer
    """
    try:
        add_node_to_buffer(analyzer, source_string, match, token_type)
    except SqlfmtBracketError:
        add_node_to_buffer(analyzer, source_string, match, fallback_token_type)


def add_comment_to_buffer(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    Create a COMMENT token from the match, then create a Comment
    from that token and append it to the Analyzer's buffer
    """
    token = Token.from_match(source_string, match, TokenType.COMMENT)
    is_standalone = (not bool(analyzer.node_buffer)) or "\n" in token.token
    comment = Comment(
        token=token,
        is_standalone=is_standalone,
        previous_node=analyzer.previous_node,
    )
    analyzer.comment_buffer.append(comment)
    analyzer.pos = token.epos


def add_jinja_comment_to_buffer(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    Create a COMMENT token from the match, then create a Comment
    from that token and append it to the Analyzer's buffer; raise
    StopRulesetLexing to revert to SQL lexing
    """
    add_comment_to_buffer(analyzer, source_string, match)
    raise StopRulesetLexing


def handle_newline(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    When a newline is encountered in the source, we typically want to create a
    new line in the Analyzer's line_buffer, flushing the node_buffer and
    comment_buffer in the process.

    However, if we have lexed a standalone comment, we do not want to create
    a Line with only that comment; instead, it must be added to the next Line
    that contains Nodes.

    That deferral is a formatting decision, so it must not apply where
    formatting is disabled: holding the comment back also swallows the newline
    that terminates it, and then every blank line that follows lands in the same
    branch and is dropped too. Flushing instead keeps the source verbatim, and
    costs nothing, because render_with_comments already skips a line whose
    content is empty once it has rendered a standalone comment.
    """
    nl_token = Token.from_match(source_string, match, TokenType.NEWLINE)
    nl_node = analyzer.node_manager.create_node(
        token=nl_token, previous_node=analyzer.previous_node
    )
    if (
        analyzer.node_buffer
        or not analyzer.comment_buffer
        or nl_node.formatting_disabled
    ):
        analyzer.node_buffer.append(nl_node)
        analyzer.line_buffer.append(
            Line.from_nodes(
                previous_node=analyzer.previous_line_node,
                nodes=analyzer.node_buffer,
                comments=analyzer.comment_buffer,
            )
        )
        analyzer.node_buffer = []
        analyzer.comment_buffer = []
    else:
        # standalone comments; don't create a line, since
        # these need to be attached to the next line with
        # contents
        pass
    analyzer.pos = nl_token.epos


def handle_semicolon(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    When we hit a semicolon, the next token may require a different rule set,
    so we need to reset the analyzer's rule stack, if new rules have been
    pushed
    """
    if analyzer.rule_stack:
        analyzer.rules = analyzer.rule_stack[0]
        analyzer.rule_stack = []

    add_node_to_buffer(
        analyzer=analyzer,
        source_string=source_string,
        match=match,
        token_type=TokenType.SEMICOLON,
    )


def handle_ddl_as(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    When we hit "as" in a create function or table statement,
    the following syntax should be parsed using the main (select) rules,
    unless the next token is a quoted name.
    """
    add_node_to_buffer(
        analyzer=analyzer,
        source_string=source_string,
        match=match,
        token_type=TokenType.UNTERM_KEYWORD,
    )

    quoted_name_rule = analyzer.get_rule("quoted_name")
    comment_rule = analyzer.get_rule("comment")

    quoted_name_pattern = rf"({comment_rule.pattern}|\s)*" + quoted_name_rule.pattern
    quoted_name_match = re.match(
        quoted_name_pattern, source_string[analyzer.pos :], re.IGNORECASE | re.DOTALL
    )

    if not quoted_name_match:
        assert analyzer.rule_stack, (
            "Internal Error! Open an issue. Could not parse DDL 'as' "
            f"at pos {analyzer.pos}. Context: "
            f"{source_string[analyzer.pos : analyzer.pos + 50]}"
        )
        analyzer.pop_rules()


def handle_closing_angle_bracket(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    When we hit ">", it could be a closing bracket, the ">" operator,
    or the first character of another operator, like ">>". We need
    to first assume it's a closing bracket, but if that raises a lexing
    error, we need to try to match the source again against the operator
    rule, to get the whole operator token
    """
    try:
        add_node_to_buffer(
            analyzer=analyzer,
            source_string=source_string,
            match=match,
            token_type=TokenType.BRACKET_CLOSE,
        )
    except SqlfmtBracketError:
        operator_rule = analyzer.get_rule("operator")
        operator_pattern = re.compile(
            r"\s*" + operator_rule.pattern,
            re.IGNORECASE | re.DOTALL,
        )
        operator_match = operator_pattern.match(source_string, analyzer.pos)

        assert operator_match, (
            "Internal Error! Open an issue. Could not parse closing bracket '>' "
            f"at pos {analyzer.pos}. Context: "
            f"{source_string[analyzer.pos : analyzer.pos + 10]}"
        )
        add_node_to_buffer(
            analyzer=analyzer,
            source_string=source_string,
            match=operator_match,
            token_type=TokenType.OPERATOR,
        )


def handle_set_operator(
    analyzer: "Analyzer", source_string: str, match: re.Match
) -> None:
    """
    Mostly, when we encounter a set operator (like union) we just want to add
    a token with a SET_OPERATOR type. However, EXCEPT is an overloaded
    keyword in some dialects (BigQuery) that support `select * except (fields)`.
    In this case, except should be a WORD_OPERATOR
    """
    previous_node = analyzer.previous_node
    token = Token.from_match(source_string, match, TokenType.SET_OPERATOR)
    prev_token, _ = get_previous_token(previous_node)
    if (
        token.token.lower() == "except"
        and prev_token
        and prev_token.type is TokenType.STAR
    ):
        token = Token(
            type=TokenType.WORD_OPERATOR,
            prefix=token.prefix,
            token=token.token,
            spos=token.spos,
            epos=token.epos,
        )
    node = analyzer.node_manager.create_node(token=token, previous_node=previous_node)
    analyzer.node_buffer.append(node)
    analyzer.pos = token.epos


def handle_number(analyzer: "Analyzer", source_string: str, match: re.Match) -> None:
    """
    We don't know if a token like "-3" or "+4" is properly a unary operator,
    or a poorly-spaced binary operator, so we have to check the previous
    node.
    """
    first_char = source_string[match.span(1)[0] : match.span(1)[0] + 1]
    if first_char in ["+", "-"] and analyzer.previous_node is not None:
        prev_token, _ = get_previous_token(analyzer.previous_node)
        if prev_token and prev_token.type in (
            TokenType.NUMBER,
            TokenType.NAME,
            TokenType.QUOTED_NAME,
            TokenType.STATEMENT_END,
            TokenType.BRACKET_CLOSE,
        ):
            # This is a binary operator. Create a new match for only the
            # operator token
            op_prog = re.compile(r"\s*(\+|-)")
            op_match = op_prog.match(source_string, pos=analyzer.pos)
            assert op_match, "Internal error! Could not match symbol of binary operator"
            add_node_to_buffer(
                analyzer=analyzer,
                source_string=source_string,
                match=op_match,
                token_type=TokenType.OPERATOR,
            )
            # we don't have to handle the rest of the number; this
            # will get called again by analyzer.lex
            return

    # in all other cases, this is a number with/out a unary operator, and we lex it
    # as a single token
    add_node_to_buffer(
        analyzer=analyzer,
        source_string=source_string,
        match=match,
        token_type=TokenType.NUMBER,
    )


def handle_reserved_keyword(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    action: Callable[["Analyzer", str, re.Match], None],
) -> None:
    """
    Reserved keywords can be used in most dialects as table or column
    names without quoting if they are qualified.
    https://github.com/tconbeer/sqlfmt/issues/599

    This checks if the previous token is a period, and if so, lexes
    this token as a name; otherwise this action executes the passed
    action (which likely adds the node as some kind of keyword).
    """
    if analyzer.previous_node is None:
        action(analyzer, source_string, match)
        return

    previous_token, _ = get_previous_token(analyzer.previous_node)
    if previous_token is not None and previous_token.type is TokenType.DOT:
        token = Token.from_match(source_string, match, token_type=TokenType.NAME)
        if not token.prefix:
            node = analyzer.node_manager.create_node(
                token=token, previous_node=analyzer.previous_node
            )
            analyzer.node_buffer.append(node)
            analyzer.pos = token.epos
            return

    # in all other cases, this is a keyword.
    action(analyzer, source_string, match)


def handle_nonreserved_top_level_keyword(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    action: Callable[["Analyzer", str, re.Match], None],
) -> None:
    """
    Checks to see if we're at depth 0 (assuming this is a name); if so, then take the
    passed action, otherwise lex it as a name.

    For example, this allows us to lex these differently:
    explain select 1;
    select explain, 1;
    """
    token = Token.from_match(source_string, match, token_type=TokenType.NAME)
    node = analyzer.node_manager.create_node(
        token=token, previous_node=analyzer.previous_node
    )
    if node.depth[0] > 0:
        analyzer.node_buffer.append(node)
        analyzer.pos = token.epos
    else:
        handle_reserved_keyword(
            analyzer=analyzer, source_string=source_string, match=match, action=action
        )


def lex_ruleset(
    analyzer: "Analyzer",
    source_string: str,
    _: re.Match,
    new_ruleset: List["Rule"],
) -> None:
    """
    Makes a nested call to analyzer.lex, with the new ruleset activated.
    """
    analyzer.push_rules(new_ruleset)
    try:
        analyzer.lex(source_string)
    except StopRulesetLexing:
        analyzer.pop_rules()


def handle_jinja_block_start(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    Lex tags like {% if ... %} and {% for ... %} that open a jinja block
    """
    add_node_to_buffer(
        analyzer=analyzer,
        source_string=source_string,
        match=match,
        token_type=TokenType.JINJA_BLOCK_START,
    )
    raise StopRulesetLexing


def handle_jinja_block_keyword(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
) -> None:
    """
    Lex tags like {% elif ... %} and {% else %} that continue an open jinja block
    """
    if analyzer.previous_node is not None:
        try:
            start_tag = analyzer.previous_node.open_jinja_blocks[-1]
        except IndexError:
            # {% if foo %}{% else %} is allowed, but then previous
            # node won't have any open jinja blocks yet.
            # when creating the node, we check to make sure these
            # match
            start_tag = analyzer.previous_node

        previous_node = start_tag.previous_node

        add_node_to_buffer(
            analyzer=analyzer,
            source_string=source_string,
            match=match,
            token_type=TokenType.JINJA_BLOCK_KEYWORD,
            previous_node=previous_node,
            override_analyzer_prev_node=True,
        )
        raise StopRulesetLexing

    else:
        raise_sqlfmt_bracket_error(analyzer, source_string, match)


def handle_jinja_data_block_start(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    new_ruleset: Optional[List[Rule]],
    raises: bool = True,
) -> None:
    """
    Lex tags like {% set foo %} and {% call my_macro %} that open a jinja block
    that can contain arbitrary data.

    This can get called from the JINJA ruleset, in which case we need to
    raise an additional StopRulesetLexing after the JINJA_DATA segment
    is fully lexed.
    """
    add_node_to_buffer(
        analyzer=analyzer,
        source_string=source_string,
        match=match,
        token_type=TokenType.JINJA_BLOCK_START,
    )
    if new_ruleset is None:
        new_ruleset = analyzer.rules
    lex_ruleset(
        analyzer,
        source_string,
        match,
        new_ruleset=new_ruleset,
    )
    if raises:
        raise StopRulesetLexing


def handle_jinja_block_end(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    reset_sql_depth: bool = False,
) -> None:
    """
    Lex tags like {% endif %} and {% endfor %} that close an open jinja block
    """
    if analyzer.previous_node is not None:
        try:
            start_tag = analyzer.previous_node.open_jinja_blocks[-1]
        except IndexError:
            # {% if foo %}{% else %} is allowed, but then previous
            # node won't have any open jinja blocks yet.
            # when creating the node, we check to make sure these
            # match
            start_tag = analyzer.previous_node

        add_node_to_buffer(
            analyzer=analyzer,
            source_string=source_string,
            match=match,
            token_type=TokenType.JINJA_BLOCK_END,
        )

        if reset_sql_depth:
            analyzer.previous_node.open_brackets = start_tag.open_brackets.copy()

        raise StopRulesetLexing

    else:
        # No open jinja blocks or none that match this token
        raise_sqlfmt_bracket_error(analyzer, source_string=source_string, match=match)


def handle_jinja(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    start_name: str,
    end_name: str,
    token_type: TokenType,
) -> None:
    """
    Lex simple jinja statements and expressions (with possibly nested curlies)
    and add to buffer
    """
    handle_potentially_nested_tokens(
        analyzer=analyzer,
        source_string=source_string,
        match=match,
        start_name=start_name,
        end_name=end_name,
        token_type=token_type,
    )
    raise StopRulesetLexing


def handle_potentially_nested_tokens(
    analyzer: "Analyzer",
    source_string: str,
    match: re.Match,
    start_name: str,
    end_name: str,
    token_type: TokenType,
) -> None:
    # extract properties from matching start of token
    """
    Lex potentially nested statements, like jinja statements or
    c-style block comments
    """
    start_rule = analyzer.get_rule(rule_name=start_name)
    end_rule = analyzer.get_rule(rule_name=end_name)

    # extract properties from matching start of token
    pos, _ = match.span(0)
    spos, epos = match.span(1)
    prefix = source_string[pos:spos]
    # construct a new regex that will match the first instance
    # of either the ending or nesting rules
    patterns = [start_rule.pattern, end_rule.pattern]
    program = re.compile(
        MAYBE_WHITESPACES + group(*patterns), re.IGNORECASE | re.DOTALL
    )
    while True:
        epos = analyzer.search_for_terminating_token(
            start_rule=start_name,
            program=program,
            nesting_program=start_rule.program,
            tail=source_string[epos:],
            pos=epos,
        )
        if start_name != "jinja_expression_start":
            break
        else:
            # jinja expressions can contain nested dictionaries whose brackets might
            # incorrectly match; e.g., {{ {'a': {'b': 1}} }}. We use the jinja lexer
            # on our match to ensure that the last token is a jinja variable_end
            # token; if it's not, we keep searching.
            jinja_tokens = list(Environment().lex(source_string[spos:epos]))
            final_token_type = jinja_tokens[-1][1]
            if final_token_type == "variable_end":
                break

    token_text = source_string[spos:epos]
    token = Token(token_type, prefix, token_text, pos, epos)
    node = analyzer.node_manager.create_node(token, analyzer.previous_node)
    analyzer.node_buffer.append(node)
    analyzer.pos = epos



from typing import List, Sequence, Tuple

from sqlfmt.exception import SqlfmtSegmentError
from sqlfmt.line import Line


def create_segments_from_lines(lines: Sequence[Line]) -> List["Segment"]:
    """
    A segment is a list of consecutive lines that are indented from the
    first line.

    This method takes a list of lines and returns a list of segments.

    Is is basically an unfold/corecursion, but due to recursion limits
    we need to do it as a loop.
    """
    segments: List["Segment"] = []
    j = 0

    while True:
        if not lines[j:]:
            return segments

        target_depth = lines[j].depth
        start_idx = j + 2 if lines[j].is_standalone_operator else j + 1
        for i, line in enumerate(lines[start_idx:], start=start_idx):
            # scan through the lines until we get back to the
            # depth of the first line
            if line.starts_new_segment(target_depth):
                segments.append(Segment(lines[j:i]))
                j = i
                break
        else:
            # we've exhausted lines without finding any segments, so append a
            # single segment comprising the original list
            segments.append(Segment(lines[j:]))
            break

    return segments


class Segment(List[Line]):
    @property
    def head(self) -> Tuple[Line, int]:
        """
        Returns the first nonblank line in the Segment, and the index
        of that line
        """
        for i, line in enumerate(self):
            if not line.is_blank_line:
                return line, i
        else:
            raise SqlfmtSegmentError("All lines in the segment are empty")

    @property
    def tail(self) -> Tuple[Line, int]:
        """
        Returns the last nonblank line in the Segment, and the index
        of that line (from the bottom. TODO: make the index more obvious)
        """
        for i, line in enumerate(reversed(self)):
            if not line.is_blank_line:
                return line, i
        else:
            raise SqlfmtSegmentError("All lines in the segment are empty")

    @property
    def tail_closes_head(self) -> bool:
        """
        Returns True only if the last line in the segment closes a bracket or
        simple jinja block that is opened by the first line in the segment.
        """
        if len(self) <= 1:
            return False

        try:
            head, i = self.head
            tail, j = self.tail
        except SqlfmtSegmentError:
            return False

        if head == tail:
            return False

        between_lines = self[i + 1 : -(j + 1)]
        if tail.depth == head.depth and (
            (
                tail.closes_bracket_from_previous_line
                and all([line.depth[0] > head.depth[0] for line in between_lines])
            )
            or (
                tail.closes_simple_jinja_block_from_previous_line
                and all([line.depth[1] > head.depth[1] for line in between_lines])
            )
        ):
            return True
        else:
            return False

    def split_after(self, idx: int) -> List["Segment"]:
        """
        Takes an index, and returns a list of either one or two segments,
        composed of the lines of self.lines[idx+1:], depending on whether the segment
        ends with a closing bracket
        """
        if self.tail_closes_head:
            _, j = self.tail
            return [
                # the lines between the head and tail
                Segment(self[idx + 1 : -(j + 1)]),
                # the tail line (and trailing whitespace)
                Segment(self[-(j + 1) :]),
            ]
        else:
            return [Segment(self[idx + 1 :])]



from abc import ABC, abstractmethod
from typing import List

from sqlfmt.analyzer import Analyzer
from sqlfmt.node_manager import NodeManager
from sqlfmt.rule import Rule
from sqlfmt.rules import MAIN


class Dialect(ABC):
    """
    Abstract class for a SQL dialect.

    Each dialect should override the RULES dict to define their own grammar. RULES
    must have a key "main" that contains the rules for the main lexing loop.
    """

    RULES: List[Rule]
    case_sensitive_names = False

    @abstractmethod
    def get_rules(self) -> List[Rule]:
        """
        Returns the Dialect's Rules, sorted by priority
        """
        return sorted(self.RULES, key=lambda rule: rule.priority)

    def initialize_analyzer(self, line_length: int) -> Analyzer:
        """
        Creates and returns an analyzer that uses the Dialect's rules for lexing
        """
        analyzer = Analyzer(
            line_length=line_length,
            rules=self.get_rules(),
            node_manager=NodeManager(self.case_sensitive_names),
        )
        return analyzer


class Polyglot(Dialect):
    """
    A universal SQL dialect meant to encompass the common usage of at least
    Postgres, MySQL, BigQuery Standard SQL, Snowflake SQL, SparkSQL.
    """

    def __init__(self) -> None:
        self.RULES = MAIN

    def get_rules(self) -> List[Rule]:
        return super().get_rules()


class ClickHouse(Polyglot):
    case_sensitive_names = True



from pathlib import Path
from typing import List

import pytest

from sqlfmt.mode import Mode
from sqlfmt.report import Report, SqlFormatResult


@pytest.fixture
def no_change_results() -> List[SqlFormatResult]:
    results = [
        SqlFormatResult(
            source_path=Path("~/path/to/file.sql"),
            source_string="select * from my_table\n",
            formatted_string="select * from my_table\n",
            encoding="utf-8",
            utf_bom="",
        ),
        SqlFormatResult(
            source_path=Path("~/path/to/another_file.sql"),
            source_string="select * from my_table where true\n",
            formatted_string="select * from my_table where true\n",
            encoding="utf-8",
            utf_bom="",
        ),
    ]
    return results


@pytest.fixture
def changed_results() -> List[SqlFormatResult]:
    results = [
        SqlFormatResult(
            source_path=Path("~/path/to/file.sql"),
            source_string="select * from my_table\n",
            formatted_string="select * from my_table\n",
            encoding="utf-8",
            utf_bom="",
        ),
        SqlFormatResult(
            source_path=Path("~/path/to/another_file.sql"),
            source_string="SELECT * from my_table where true",
            formatted_string="select * from my_table where true\n",
            encoding="utf-8",
            utf_bom="",
        ),
        SqlFormatResult(
            source_path=Path("~/path/to/yet_another_file.sql"),
            source_string="select a,\n b\n * from my_table where \n a = b\n",
            formatted_string="select a, b from my_table where a = b\n",
            encoding="utf-8",
            utf_bom="",
        ),
    ]
    return results


def test_no_change_report(
    no_change_results: List[SqlFormatResult], default_mode: Mode
) -> None:
    report = Report(no_change_results, default_mode)
    assert report
    assert str(report) == "2 files left unchanged."


def test_no_change_verbose_report(
    no_change_results: List[SqlFormatResult], verbose_mode: Mode
) -> None:
    report = Report(no_change_results, verbose_mode)
    assert report

    expected_report = (
        "2 files left unchanged.\n"
        f"{Path('~/path/to/another_file.sql')} left unchanged.\n"
        f"{Path('~/path/to/file.sql')} left unchanged."
    )
    assert str(report) == expected_report


def test_changed_report_default_mode(
    changed_results: List[SqlFormatResult], default_mode: Mode
) -> None:
    report = Report(changed_results, default_mode)
    assert report
    expected_report = (
        "\x1b[1m2 files formatted.\x1b[0m\n"
        "1 file left unchanged.\n"
        f"{Path('~/path/to/another_file.sql')} formatted.\n"
        f"{Path('~/path/to/yet_another_file.sql')} formatted."
    )
    assert str(report) == expected_report


def test_changed_report_verbose_mode(
    changed_results: List[SqlFormatResult], verbose_mode: Mode
) -> None:
    report = Report(changed_results, verbose_mode)
    assert report
    expected_report = (
        "\x1b[1m2 files formatted.\x1b[0m\n"
        "1 file left unchanged.\n"
        f"{Path('~/path/to/another_file.sql')} formatted.\n"
        f"{Path('~/path/to/yet_another_file.sql')} formatted.\n"
        f"{Path('~/path/to/file.sql')} left unchanged."
    )
    assert str(report) == expected_report


def test_changed_report_check_mode(
    changed_results: List[SqlFormatResult], check_mode: Mode
) -> None:
    report = Report(changed_results, check_mode)
    assert report
    expected_report = (
        "\x1b[1m2 files failed formatting check.\x1b[0m\n"
        "1 file passed formatting check.\n"
        f"{Path('~/path/to/another_file.sql')} failed formatting check.\n"
        f"{Path('~/path/to/yet_another_file.sql')} failed formatting check."
    )
    assert str(report) == expected_report


def test_changed_report_verbose_check_mode(
    changed_results: List[SqlFormatResult], verbose_check_mode: Mode
) -> None:
    report = Report(changed_results, verbose_check_mode)
    assert report
    expected_report = (
        "\x1b[1m2 files failed formatting check.\x1b[0m\n"
        "1 file passed formatting check.\n"
        f"{Path('~/path/to/another_file.sql')} failed formatting check.\n"
        f"{Path('~/path/to/yet_another_file.sql')} failed formatting check.\n"
        f"{Path('~/path/to/file.sql')} passed formatting check."
    )
    assert str(report) == expected_report


def test_no_change_report_check_mode(
    no_change_results: List[SqlFormatResult], check_mode: Mode
) -> None:
    report = Report(no_change_results, check_mode)
    assert report
    assert str(report) == "2 files passed formatting check."


def test_no_change_report_diff_mode(
    no_change_results: List[SqlFormatResult], diff_mode: Mode
) -> None:
    report = Report(no_change_results, diff_mode)
    assert report
    assert str(report) == "2 files would be left unchanged."


def test_changed_report_diff_mode(
    changed_results: List[SqlFormatResult], diff_mode: Mode
) -> None:
    report = Report(changed_results, diff_mode)
    expected_report = (
        "\x1b[1m2 files would be formatted.\x1b[0m\n"
        "1 file would be left unchanged.\n"
        f"{Path('~/path/to/another_file.sql')} would be formatted.\n"
        f"{Path('~/path/to/yet_another_file.sql')} would be formatted.\n"
        f"\x1b[31m\x1b[22m--- {Path('~/path/to/another_file.sql')}\t(Source Query)\n"
        f"\x1b[0m\x1b[32m\x1b[22m+++ {Path('~/path/to/another_file.sql')}\t(Formatted Query)\n"  # noqa: E501
        "\x1b[0m\x1b[36m\x1b[22m@@ -1 +1 @@\n"
        "\x1b[0m\x1b[31m\x1b[22m-SELECT * from my_table where true\n"
        "\x1b[0m\\ No newline at end of file\n"
        "\x1b[32m\x1b[22m+select * from my_table where true\n"
        "\x1b[0m\n"
        f"\x1b[31m\x1b[22m--- {Path('~/path/to/yet_another_file.sql')}\t(Source Query)\n"  # noqa: E501
        f"\x1b[0m\x1b[32m\x1b[22m+++ {Path('~/path/to/yet_another_file.sql')}\t(Formatted Query)\n"  # noqa: E501
        "\x1b[0m\x1b[36m\x1b[22m@@ -1,4 +1 @@\n"
        "\x1b[0m\x1b[31m\x1b[22m-select a,\n"
        "\x1b[0m\x1b[31m\x1b[22m- b\n"
        "\x1b[0m\x1b[31m\x1b[22m- * from my_table where \n"
        "\x1b[0m\x1b[31m\x1b[22m- a = b\n"
        "\x1b[0m\x1b[32m\x1b[22m+select a, b from my_table where a = b\n"
        "\x1b[0m"
    )
    assert report
    assert str(report) == expected_report


def test_changed_report_no_color_diff_mode(
    changed_results: List[SqlFormatResult], no_color_diff_mode: Mode
) -> None:
    report = Report(changed_results, no_color_diff_mode)
    expected_report = (
        "2 files would be formatted.\n"
        "1 file would be left unchanged.\n"
        f"{Path('~/path/to/another_file.sql')} would be formatted.\n"
        f"{Path('~/path/to/yet_another_file.sql')} would be formatted.\n"
        f"--- {Path('~/path/to/another_file.sql')}\t(Source Query)\n"
        f"+++ {Path('~/path/to/another_file.sql')}\t(Formatted Query)\n"
        "@@ -1 +1 @@\n"
        "-SELECT * from my_table where true\n"
        "\\ No newline at end of file\n"
        "+select * from my_table where true\n"
        "\n"
        f"--- {Path('~/path/to/yet_another_file.sql')}\t(Source Query)\n"
        f"+++ {Path('~/path/to/yet_another_file.sql')}\t(Formatted Query)\n"
        "@@ -1,4 +1 @@\n"
        "-select a,\n"
        "- b\n"
        "- * from my_table where \n"
        "- a = b\n"
        "+select a, b from my_table where a = b\n"
        ""
    )
    assert report
    assert str(report) == expected_report



from dataclasses import dataclass
from typing import List, Optional

from sqlfmt.jinjafmt import JinjaFormatter
from sqlfmt.line import Line
from sqlfmt.merger import LineMerger
from sqlfmt.mode import Mode
from sqlfmt.node import Node
from sqlfmt.node_manager import NodeManager
from sqlfmt.query import Query
from sqlfmt.splitter import LineSplitter


@dataclass
class QueryFormatter:
    mode: Mode

    def _split_lines(self, lines: List[Line]) -> List[Line]:
        """
        Splits lines to make line depth consistent and syntax
        apparent
        """
        node_manager = NodeManager(self.mode.dialect.case_sensitive_names)
        splitter = LineSplitter(node_manager)
        new_lines = []
        for line in lines:
            splits = list(splitter.maybe_split(line))
            new_lines.extend(splits)
        return new_lines

    def _format_jinja(self, lines: List[Line]) -> List[Line]:
        """
        Formats the contents of jinja tags (the code between
        the curlies) by mutating existing jinja nodes
        """
        formatter = JinjaFormatter(mode=self.mode)
        new_lines: List[Line] = []
        for line in lines:
            new_lines.extend(formatter.format_line(line))
        return new_lines

    def _merge_lines(self, lines: List[Line]) -> List[Line]:
        """
        Merge lines to minimize vertical space used by the
        query, while maintaining the syntax hierarchy achieved
        by the splitter
        """
        merger = LineMerger(mode=self.mode)
        lines = merger.maybe_merge_lines(lines)
        return lines

    def _dedent_jinja_blocks(self, lines: List[Line]) -> List[Line]:
        """
        Jinja block tags, like {% if foo %} and {% endif %}, shouldn't
        be printed at their depth, since their contents may be dedented
        farther. This dedents the tags as necessary, in a single pass
        """
        start_node: Optional[Node] = None
        for line in lines:
            if (
                line.is_standalone_jinja_statement
                and line.nodes[0].is_closing_jinja_block
                and not line.formatting_disabled
            ):
                assert start_node
                line.nodes[0].open_brackets = start_node.open_brackets

            if line.nodes and line.nodes[-1].open_jinja_blocks:
                start_node = line.nodes[-1].open_jinja_blocks[-1]
                if (
                    len(line.open_brackets) < len(start_node.open_brackets)
                    and not start_node.formatting_disabled
                ):
                    start_node.open_brackets = line.open_brackets

        return lines

    def _remove_extra_blank_lines(self, lines: List[Line]) -> List[Line]:
        """
        A query can have at most 2 consecutive blank lines at depth (0,0)
        and 1 consecutive blank line at any other depth. See issue #249
        for motivation and details.
        """
        new_lines: List[Line] = []
        # initialize cnt high so we remove any extra lines at the beginning
        # of files.
        cnt = 2
        for line in lines:
            if line.is_blank_line:
                max_cnt = 2 if line.depth == (0, 0) else 1
                if cnt < max_cnt or line.formatting_disabled:
                    new_lines.append(line)
                cnt += 1
            else:
                new_lines.append(line)
                cnt = 0
        return new_lines

    def format(self, raw_query: Query) -> Query:
        """
        Applies 4 transformations to a Query:
        1. Splits lines
        2. Formats jinja tags
        3. Dedents jinja block tags to match their least-indented contents
        4. Merges lines
        5. Removes extra blank lines
        """
        lines = raw_query.lines

        pipeline = [
            self._split_lines,
            self._format_jinja,
            self._dedent_jinja_blocks,
            self._merge_lines,
            self._remove_extra_blank_lines,
        ]

        for transform in pipeline:
            lines = transform(lines)

        formatted_query = Query(
            source_string=raw_query.source_string,
            line_length=raw_query.line_length,
            lines=lines,
        )

        return formatted_query



from dataclasses import dataclass, field
from typing import List

from sqlfmt.line import Line
from sqlfmt.node import Node
from sqlfmt.tokens import Token


@dataclass
class Query:
    """
    A Query is a collection of Lines, the corresponding raw source string, and
    the desired line length. Queries are mutated by the Formatter
    """

    source_string: str
    line_length: int
    lines: List[Line] = field(default_factory=list)

    @property
    def tokens(self) -> List[Token]:
        tokens: List[Token] = []
        for line in self.lines:
            tokens.extend(line.tokens)
        return tokens

    @property
    def nodes(self) -> List[Node]:
        nodes: List[Node] = []
        for line in self.lines:
            nodes.extend(line.nodes)
        return nodes

    def __str__(self) -> str:
        draft = [line.render_with_comments(self.line_length) for line in self.lines]
        return "".join(draft)



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import ALTER_DROP_FUNCTION, CREATE_FUNCTION, group
from sqlfmt.rules.core import CORE
from sqlfmt.tokens import TokenType

FUNCTION = [
    *CORE,
    Rule(
        name="function_as",
        priority=1100,
        pattern=group(
            r"as",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=actions.handle_ddl_as,
        ),
    ),
    Rule(
        name="word_operator",
        priority=1200,
        pattern=group(
            r"to",
            r"from",
            # snowflake
            r"runtime_version",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.WORD_OPERATOR
            ),
        ),
    ),
    Rule(
        name="unterm_keyword",
        priority=1300,
        pattern=group(
            CREATE_FUNCTION,
            ALTER_DROP_FUNCTION,
            r"language",
            r"transform",
            r"immutable",
            r"stable",
            r"volatile",
            r"(not\s+)?leakproof",
            r"volatile",
            r"called\s+on\s+null\s+input",
            r"returns\s+null\s+on\s+null\s+input",
            r"return(s)?(?!\s+null)",
            r"strict",
            r"(external\s+)?security\s+(invoker|definer)",
            r"parallel\s+(unsafe|restricted|safe)",
            r"cost",
            r"rows",
            r"support",
            # snowflake
            r"((un)?set\s+)?comment",
            r"imports",
            r"packages",
            r"handler",
            r"target_path",
            r"(not\s+)?null",
            # snowflake external functions
            r"((un)?set\s+)?"
            + group(
                r"api_integration",
                r"headers",
                r"context_headers",
                r"max_batch_rows",
                r"compression",
                r"request_translator",
                r"response_translator",
            ),
            # bq
            r"options",
            r"remote\s+with\s+connection",
            # ALTER
            r"rename\s+to",
            r"owner\s+to",
            r"set\s+schema",
            r"(no\s+)?depends\s+on\s+extension",
            r"cascade",
            r"restrict",
            # alter snowflake
            r"(un)?set\s+secure",
            # pg catchall for set
            r"(re)?set(\s+all)?",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
]



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import NEWLINE, group
from sqlfmt.tokens import TokenType

JINJA_SHARED_PATTERNS = {
    "set": r"\{%-?\s*set\s+[^=]+?-?%\}",
    "endset": r"\{%-?\s*endset\s*-?%\}",
    "call": r"\{%-?\s*call(\(.*?\))?\s+\w+.*?-?%\}",
    "endcall": r"\{%-?\s*endcall\s*-?%\}",
}

JINJA_DATA = [
    Rule(
        name="jinja_set_block_start",
        priority=100,
        pattern=group(JINJA_SHARED_PATTERNS["set"]),
        action=partial(
            actions.handle_jinja_data_block_start,
            new_ruleset=None,
            raises=False,
        ),
    ),
    Rule(
        name="jinja_set_block_end",
        priority=101,
        pattern=group(JINJA_SHARED_PATTERNS["endset"]),
        action=partial(
            actions.handle_jinja_block_end,
            reset_sql_depth=True,
        ),
    ),
    Rule(
        name="jinja_call_block_start",
        priority=261,
        pattern=group(JINJA_SHARED_PATTERNS["call"]),
        action=partial(
            actions.handle_jinja_data_block_start,
            new_ruleset=None,
            raises=False,
        ),
    ),
    Rule(
        name="jinja_call_block_end",
        priority=265,
        pattern=group(JINJA_SHARED_PATTERNS["endcall"]),
        action=partial(
            actions.handle_jinja_block_end,
            reset_sql_depth=True,
        ),
    ),
    Rule(
        name="jinja_newline",
        priority=500,
        pattern=group(NEWLINE),
        action=actions.handle_newline,
    ),
    Rule(
        name="jinja_data",
        priority=600,
        pattern=group(r".*?") + r"\s*" + group(*JINJA_SHARED_PATTERNS.values(), r"$"),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.DATA),
    ),
]

JINJA = [
    Rule(
        name="jinja_comment",
        priority=0,
        pattern=group(r"\{\#.*?\#\}"),
        action=actions.add_jinja_comment_to_buffer,
    ),
    Rule(
        name="jinja_set_block_start",
        priority=100,
        pattern=group(JINJA_SHARED_PATTERNS["set"]),
        action=partial(
            actions.handle_jinja_data_block_start,
            new_ruleset=JINJA_DATA,
        ),
    ),
    Rule(
        name="jinja_set_block_end",
        priority=101,
        pattern=group(JINJA_SHARED_PATTERNS["endset"]),
        action=actions.raise_sqlfmt_bracket_error,
    ),
    Rule(
        name="jinja_if_block_start",
        priority=200,
        pattern=group(r"\{%-?\s*if.*?-?%\}"),
        action=actions.handle_jinja_block_start,
    ),
    Rule(
        name="jinja_elif_block_start",
        priority=201,
        pattern=group(r"\{%-?\s*elif\s+\w+.*?-?%\}"),
        action=actions.handle_jinja_block_keyword,
    ),
    Rule(
        name="jinja_else_block_start",
        priority=202,
        pattern=group(r"\{%-?\s*else\s*-?%\}"),
        action=actions.handle_jinja_block_keyword,
    ),
    Rule(
        name="jinja_if_block_end",
        priority=203,
        pattern=group(r"\{%-?\s*endif\s*-?%\}"),
        action=actions.handle_jinja_block_end,
    ),
    Rule(
        name="jinja_for_block_start",
        priority=210,
        pattern=group(r"\{%-?\s*for\s+.*?-?%\}"),
        action=actions.handle_jinja_block_start,
    ),
    Rule(
        name="jinja_for_block_end",
        priority=211,
        pattern=group(r"\{%-?\s*endfor\s*-?%\}"),
        action=actions.handle_jinja_block_end,
    ),
    Rule(
        name="jinja_macro_block_start",
        priority=220,
        pattern=group(r"\{%-?\s*macro\s+\w+.*?-?%\}"),
        action=actions.handle_jinja_block_start,
    ),
    Rule(
        name="jinja_macro_block_end",
        priority=221,
        pattern=group(r"\{%-?\s*endmacro\s*-?%\}"),
        action=partial(
            actions.handle_jinja_block_end,
            reset_sql_depth=True,
        ),
    ),
    Rule(
        name="jinja_test_block_start",
        priority=230,
        pattern=group(r"\{%-?\s*test\s+\w+.*?-?%\}"),
        action=actions.handle_jinja_block_start,
    ),
    Rule(
        name="jinja_test_block_end",
        priority=231,
        pattern=group(r"\{%-?\s*endtest\s*-?%\}"),
        action=partial(
            actions.handle_jinja_block_end,
            reset_sql_depth=True,
        ),
    ),
    Rule(
        name="jinja_snapshot_block_start",
        priority=240,
        pattern=group(r"\{%-?\s*snapshot\s+\w+.*?-?%\}"),
        action=actions.handle_jinja_block_start,
    ),
    Rule(
        name="jinja_snapshot_block_end",
        priority=241,
        pattern=group(r"\{%-?\s*endsnapshot\s*-?%\}"),
        action=partial(
            actions.handle_jinja_block_end,
            reset_sql_depth=True,
        ),
    ),
    Rule(
        name="jinja_materialization_block_start",
        priority=250,
        pattern=group(r"\{%-?\s*materialization\s+\w+\s*,.*?-?%\}"),
        action=actions.handle_jinja_block_start,
    ),
    Rule(
        name="jinja_materialization_block_end",
        priority=251,
        pattern=group(r"\{%-?\s*endmaterialization\s*-?%\}"),
        action=partial(
            actions.handle_jinja_block_end,
            reset_sql_depth=True,
        ),
    ),
    # call blocks that are used to call dbt's statement macro
    # are guaranteed to contain SQL, so we can parse them
    # like ordinary jinja blocks, and format the contents
    Rule(
        name="jinja_call_statement_block_start",
        priority=260,
        pattern=group(r"\{%-?\s*call\s+(noop_)?statement\(.*?\)\s*-?%\}"),
        action=actions.handle_jinja_block_start,
    ),
    # call blocks that call other macros may contain SQL or
    # arbitrary text or data, so we need to parse the whole block
    # as DATA so we don't format it
    Rule(
        name="jinja_call_block_start",
        priority=261,
        pattern=group(JINJA_SHARED_PATTERNS["call"]),
        action=partial(
            actions.handle_jinja_data_block_start,
            new_ruleset=JINJA_DATA,
        ),
    ),
    Rule(
        name="jinja_call_block_end",
        priority=265,
        pattern=group(JINJA_SHARED_PATTERNS["endcall"]),
        action=partial(
            actions.handle_jinja_block_end,
            reset_sql_depth=True,
        ),
    ),
    Rule(
        name="jinja_statement_start",
        priority=500,
        pattern=group(r"\{%-?"),
        action=partial(
            actions.handle_jinja,
            start_name="jinja_statement_start",
            end_name="jinja_statement_end",
            token_type=TokenType.JINJA_STATEMENT,
        ),
    ),
    Rule(
        name="jinja_expression_start",
        priority=510,
        pattern=group(r"\{\{-?"),
        action=partial(
            actions.handle_jinja,
            start_name="jinja_expression_start",
            end_name="jinja_expression_end",
            token_type=TokenType.JINJA_EXPRESSION,
        ),
    ),
    Rule(
        name="jinja_statement_end",
        priority=600,
        pattern=group(r"-?%\}"),
        action=actions.raise_sqlfmt_bracket_error,
    ),
    Rule(
        name="jinja_expression_end",
        priority=610,
        pattern=group(r"-?\}\}"),
        action=actions.raise_sqlfmt_bracket_error,
    ),
]



from copy import deepcopy
from typing import List

import pytest

from sqlfmt.analyzer import Analyzer
from sqlfmt.comment import Comment
from sqlfmt.line import Line
from sqlfmt.mode import Mode
from sqlfmt.node_manager import NodeManager
from sqlfmt.tokens import Token, TokenType
from tests.util import read_test_data


@pytest.fixture
def node_manager(default_mode: Mode) -> NodeManager:
    return NodeManager(default_mode.dialect.case_sensitive_names)


@pytest.fixture
def source_string() -> str:
    return "with abc as (select * from my_table)\n"


@pytest.fixture
def bare_line() -> Line:
    line = Line(previous_node=None)
    return line


@pytest.fixture
def tokens() -> List[Token]:
    tokens = [
        Token(type=TokenType.UNTERM_KEYWORD, prefix="", token="with", spos=0, epos=4),
        Token(type=TokenType.NAME, prefix=" ", token="abc", spos=4, epos=8),
        Token(type=TokenType.WORD_OPERATOR, prefix=" ", token="as", spos=8, epos=11),
        Token(type=TokenType.BRACKET_OPEN, prefix=" ", token="(", spos=11, epos=13),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix="", token="select", spos=13, epos=19
        ),
        Token(type=TokenType.STAR, prefix=" ", token="*", spos=19, epos=21),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix=" ", token="from", spos=21, epos=26
        ),
        Token(type=TokenType.NAME, prefix=" ", token="my_table", spos=26, epos=35),
        Token(type=TokenType.BRACKET_CLOSE, prefix="", token=")", spos=35, epos=36),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=36, epos=36),
    ]
    return tokens


@pytest.fixture
def simple_line(
    tokens: List[Token], bare_line: Line, node_manager: NodeManager
) -> Line:
    simple_line = deepcopy(bare_line)
    previous_node = None
    for token in tokens:
        node = node_manager.create_node(token, previous_node)
        simple_line.nodes.append(node)
        previous_node = node
    return simple_line


def test_bare_line(bare_line: Line) -> None:
    assert str(bare_line) == ""

    assert not bare_line.starts_with_unterm_keyword
    assert not bare_line.contains_unterm_keyword
    assert not bare_line.contains_operator
    assert not bare_line.starts_with_comma
    assert not bare_line.starts_with_bracket_operator
    assert not bare_line.starts_with_operator
    assert not bare_line.is_blank_line
    assert not bare_line.is_too_long(88)
    assert not bare_line.opens_new_bracket
    assert not bare_line.contains_jinja
    assert bare_line.open_brackets == []
    assert bare_line.open_jinja_blocks == []


def test_simple_line(
    source_string: str, tokens: List[Token], simple_line: Line
) -> None:
    assert simple_line.depth == (0, 0)
    assert len(simple_line.nodes) == len(tokens)
    assert simple_line.open_brackets == []

    assert str(simple_line) == source_string

    expected_token_repr = (
        "Token(type=TokenType.UNTERM_KEYWORD, prefix='', token='with', spos=0, epos=4)"
    )
    assert repr(simple_line.tokens[0]) == expected_token_repr
    new_token = eval(repr(simple_line.tokens[0]))
    assert simple_line.tokens[0] == new_token

    expected_node_repr = (
        "Node(\n"
        "\ttoken='Token(type=TokenType.UNTERM_KEYWORD, token=with, spos=0)',\n"
        "\tprevious_node=None,\n"
        "\tdepth=(0, 0),\n"
        "\tprefix=' ',\n"
        "\tvalue='with',\n"
        "\topen_brackets=[],\n"
        "\topen_jinja_blocks=[],\n"
        "\tformatting_disabled=[]\n"
        ")"
    )
    assert repr(simple_line.nodes[0]) == expected_node_repr

    assert simple_line.starts_with_unterm_keyword
    assert simple_line.contains_unterm_keyword
    assert simple_line.contains_operator
    assert not simple_line.starts_with_comma
    assert not simple_line.starts_with_bracket_operator
    assert not simple_line.starts_with_operator
    assert not simple_line.is_blank_line
    assert not simple_line.is_standalone_comma
    assert not simple_line.is_too_long(88)
    assert not simple_line.opens_new_bracket
    assert not simple_line.previous_line_has_open_jinja_blocks_not_keywords
    assert not simple_line.contains_jinja

    assert simple_line.nodes[5].token.type is TokenType.STAR
    assert not simple_line.nodes[5].is_multiplication_star


def test_bare_append_newline(bare_line: Line, node_manager: NodeManager) -> None:
    # this line has no nodes
    assert not bare_line.nodes
    assert not bare_line.previous_node

    node_manager.append_newline(bare_line)
    assert bare_line.nodes
    assert bare_line.is_blank_line
    assert str(bare_line) == "\n"
    new_last_node = bare_line.nodes[-1]
    assert new_last_node.token.type is TokenType.NEWLINE
    assert (new_last_node.token.spos, new_last_node.token.epos) == (0, 0)


def test_bare_with_previous_open_lists(bare_line: Line, simple_line: Line) -> None:
    bare_line.previous_node = simple_line.nodes[-1]
    assert bare_line.open_jinja_blocks == simple_line.nodes[-1].open_jinja_blocks
    assert bare_line.open_brackets == simple_line.nodes[-1].open_brackets


def test_bare_with_previous_append_newline(
    bare_line: Line, simple_line: Line, node_manager: NodeManager
) -> None:
    bare_line.previous_node = simple_line.nodes[-1]
    node_manager.append_newline(bare_line)
    assert bare_line.nodes
    new_last_node = bare_line.nodes[-1]
    previous_token = simple_line.nodes[-1].token
    expected_position = (
        previous_token.epos,
        previous_token.epos,
    )
    assert (new_last_node.token.spos, new_last_node.token.epos) == expected_position


def test_simple_append_newline(simple_line: Line, node_manager: NodeManager) -> None:
    # this line already ends with a newline
    last_node = simple_line.nodes[-1]
    assert last_node.token.type is TokenType.NEWLINE
    assert last_node.previous_node is not None
    assert last_node.previous_node.token.type is not TokenType.NEWLINE

    node_manager.append_newline(simple_line)
    new_last_node = simple_line.nodes[-1]
    assert new_last_node != last_node
    assert new_last_node.token.type is TokenType.NEWLINE
    assert new_last_node.previous_node == last_node
    assert new_last_node.previous_node.token == last_node.token


@pytest.mark.parametrize(
    "raw_comment,normalized_comment",
    [
        ("-- my comment", "-- my comment"),
        ("--my comment", "-- my comment"),
        ("--    my comment", "-- my comment"),
        ("# my comment", "# my comment"),
        ("#my comment", "# my comment"),
        ("#    my comment", "# my comment"),
        ("/* my comment */", "/* my comment */"),
        ("/*my comment*/", "/* my comment*/"),
        ("/*    my comment */", "/* my comment */"),
    ],
)
def test_comment_rendering(
    simple_line: Line,
    bare_line: Line,
    raw_comment: str,
    normalized_comment: str,
    node_manager: NodeManager,
) -> None:
    assert simple_line.render_with_comments(88) == str(simple_line)
    assert bare_line.render_with_comments(88) == str(bare_line)

    last_node = simple_line.nodes[-1]

    comment_token = Token(
        type=TokenType.COMMENT,
        prefix="",
        token=raw_comment,
        spos=last_node.token.epos,
        epos=last_node.token.epos + 13,
    )

    inline_comment = Comment(
        token=comment_token, is_standalone=False, previous_node=last_node
    )
    simple_line.comments = [inline_comment]
    expected_inline_render = (
        str(simple_line).rstrip() + "  " + normalized_comment + "\n"
    )
    assert simple_line.render_with_comments(88) == expected_inline_render

    standalone_comment = Comment(
        token=comment_token, is_standalone=True, previous_node=None
    )
    simple_line.comments = [standalone_comment]
    expected_standalone_render = (
        simple_line.prefix + normalized_comment + "\n" + str(simple_line)
    )
    assert simple_line.render_with_comments(88) == expected_standalone_render


def test_long_comment_wrapping(simple_line: Line) -> None:
    last_node = simple_line.nodes[-1]
    comment_token = Token(
        type=TokenType.COMMENT,
        prefix="",
        token="-- " + ("asdf " * 20),
        spos=last_node.token.epos,
        epos=last_node.token.epos + 13,
    )
    comment = Comment(token=comment_token, is_standalone=True, previous_node=None)
    simple_line.comments = [comment]
    expected_render = (
        "-- asdf asdf asdf asdf asdf asdf asdf asdf asdf asdf asdf asdf asdf asdf "
        "asdf asdf\n"
        "-- asdf asdf asdf asdf\n"
        "with abc as (select * from my_table)\n"
    )
    assert simple_line.render_with_comments(88) == expected_render


@pytest.mark.parametrize(
    "raw_comment",
    [
        "-- " + ("asdf" * 30),
        "{# " + ("asdf" * 30) + "#}",
        "/* " + ("asdf" * 30) + "*/",
    ],
)
def test_long_comments_that_are_not_wrapped(
    simple_line: Line, raw_comment: str
) -> None:
    last_node = simple_line.nodes[-1]
    comment_token = Token(
        type=TokenType.COMMENT,
        prefix="",
        token=raw_comment,
        spos=last_node.token.epos,
        epos=last_node.token.epos + 13,
    )
    comment = Comment(token=comment_token, is_standalone=True, previous_node=None)
    simple_line.comments = [comment]
    expected_render = raw_comment + "\nwith abc as (select * from my_table)\n"
    assert simple_line.render_with_comments(88) == expected_render


def test_is_standalone_jinja_statement(
    bare_line: Line, simple_line: Line, node_manager: NodeManager
) -> None:
    assert not bare_line.is_standalone_jinja_statement
    assert not simple_line.is_standalone_jinja_statement

    jinja_statement = Token(
        type=TokenType.JINJA_STATEMENT,
        prefix="",
        token="{% my JINJA %}",
        spos=0,
        epos=14,
    )

    bare_line.nodes.append(node_manager.create_node(jinja_statement, None))
    simple_line.nodes.append(
        node_manager.create_node(jinja_statement, simple_line.nodes[-1])
    )

    assert bare_line.is_standalone_jinja_statement
    assert not simple_line.is_standalone_jinja_statement

    node_manager.append_newline(bare_line)
    node_manager.append_newline(simple_line)

    assert bare_line.is_standalone_jinja_statement
    assert not simple_line.is_standalone_jinja_statement


def test_closes_bracket_from_previous_line(
    simple_line: Line, default_mode: Mode
) -> None:
    assert not simple_line.closes_bracket_from_previous_line

    source_string = (
        "case\n"
        "    when\n"
        "        (\n"
        "            field_one\n"
        "            + (field_two)\n"
        "            + field_three\n"
        "        )\n"
        "    then true\n"
        "end\n"
    )
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    result = [line.closes_bracket_from_previous_line for line in q.lines]
    expected = [False, False, False, False, False, False, True, False, True]
    assert result == expected

    opens_result = [line.opens_new_bracket for line in q.lines]
    opens_expected = [True, False, True, False, False, False, False, False, False]
    assert opens_result == opens_expected


def test_jinja_whitespace_after_newline(default_mode: Mode) -> None:
    source_string = "select *\nfrom {{model}}\nwhere {{ column_name }} < 0\n"
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    where_node = [node for node in q.nodes if node.value == "where"][0]
    assert where_node.prefix == " "  # one space


def test_formatting_disabled(default_mode: Mode) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_line/test_formatting_disabled.sql"
    )
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    expected = [
        False,  # select
        True,  # -- fmt: off
        True,  # 1, 2, 3
        True,  # 4, 5, 6
        True,  # -- fmt: on
        False,  # from something
        True,  # join something_else -- fmt: off
        True,  # --fmt: on
        False,  # where format is true
    ]
    actual = [bool(line.formatting_disabled) for line in q.lines]
    assert actual == expected


def test_is_multiplication_star_bare_line(
    bare_line: Line, node_manager: NodeManager
) -> None:
    star = Token(
        type=TokenType.STAR,
        prefix="",
        token="*",
        spos=0,
        epos=1,
    )
    star_node = node_manager.create_node(star, None)
    bare_line.nodes.append(star_node)
    assert bare_line.nodes[0].token == star
    assert not bare_line.nodes[0].is_multiplication_star


def test_is_standalone_operator(
    bare_line: Line, simple_line: Line, node_manager: NodeManager
) -> None:
    assert not bare_line.is_standalone_operator
    assert not simple_line.is_standalone_operator

    plus = Token(
        type=TokenType.OPERATOR,
        prefix="",
        token="+",
        spos=0,
        epos=1,
    )
    plus_node = node_manager.create_node(plus, None)
    bare_line.nodes.append(plus_node)
    assert bare_line.is_standalone_operator
    node_manager.append_newline(bare_line)
    assert bare_line.is_standalone_operator


def test_semicolon_resets_depth(simple_line: Line, node_manager: NodeManager) -> None:
    semicolon = Token(
        type=TokenType.SEMICOLON,
        prefix="",
        token=";",
        spos=36,
        epos=37,
    )
    semicolon_node = node_manager.create_node(semicolon, simple_line.nodes[-1])
    assert simple_line.nodes[-1].depth[0] > 0
    simple_line.nodes.append(semicolon_node)
    assert simple_line.nodes[-1].depth[0] == 0


def test_previous_line_has_open_jinja_blocks_not_keywords(
    default_analyzer: Analyzer,
) -> None:
    source_string = """
    {% set a = 1 %}
    {% if a == 1 %}
    select 'foo'
    {% else %}
    select 'bar'
    {% endif %}
    """
    q = default_analyzer.parse_query(source_string=source_string)

    expected = [
        False,
        False,
        False,
        True,
        True,
        False,
        False,
    ]
    actual = [line.previous_line_has_open_jinja_blocks_not_keywords for line in q.lines]
    assert actual == expected


def test_closes_simple_jinja_block_from_previous_line(
    default_analyzer: Analyzer,
) -> None:
    source_string = """
    {% if a == 1 %}
    select 'foo'
    {% else %}
    select 'bar'
    {% endif %}
    {% if b == 2 -%} select 'baz'
    {% endif %}
    """
    q = default_analyzer.parse_query(source_string=source_string)

    expected = [False] * (len(q.lines) - 1) + [True]
    actual = [line.closes_simple_jinja_block_from_previous_line for line in q.lines]
    assert actual == expected


def test_formatting_disabled_to_str(
    default_analyzer: Analyzer,
) -> None:
    source_string = "SELECT      FOO   ,   BAR"
    q = default_analyzer.parse_query(source_string=source_string)
    assert len(q.lines) == 1
    q.lines[0].formatting_disabled = [q.nodes[0].token]

    assert str(q.lines[0]) == source_string + "\n"


def test_line_is_too_long(
    default_analyzer: Analyzer,
) -> None:
    max_length = 88
    source_string = "a" * max_length + "\n" + "b" * (max_length + 1) + "\n"
    q = default_analyzer.parse_query(source_string=source_string)
    assert len(q.lines) == 2
    assert not q.lines[0].is_too_long(max_length=max_length)
    assert q.lines[1].is_too_long(max_length=max_length)



import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from sqlfmt.analyzer import Analyzer

MAYBE_WHITESPACES: str = r"[^\S\n]*"  # any whitespace except newline


@dataclass
class Rule:
    """
    A lex rule.

    When the analyzer lexes the source string, it applies each Rule in turn,
    in ascending order of priority.

    The Analyzer tries to match the Rule's regex pattern to the current position
    in the source string. If there is a match, the Analyzer calls the Rule's
    action with the arguments (self [the analyzer], source_string, match). The
    Rule's action may mutate the analyzer's buffer (if desired). The action
    must return the position in the source_string where the Analyzer should
    look for the next match.
    """

    name: str
    priority: int  # lower get matched first
    pattern: str
    action: Callable[["Analyzer", str, re.Match], None]

    def __post_init__(self) -> None:
        self.program = re.compile(
            MAYBE_WHITESPACES + self.pattern, re.IGNORECASE | re.DOTALL
        )



from pathlib import Path
from typing import Any, List

import pytest

from sqlfmt.config import _find_config_file, _get_common_parents, _load_config_from_path
from sqlfmt.exception import SqlfmtConfigError
from tests.util import copy_config_file_to_dst


@pytest.fixture(
    params=[
        [""],
        ["subdir"],
        ["subdir", "another"],
        ["model.sql"],
        ["model.sql", "another.sql"],
        ["subdir/model.sql"],
        ["subdir/model.sql", "another_dir"],
        ["subdir/another_dir/yet_another_dir/"],
        ["subdir/another_dir/yet_another_dir/model.sql"],
    ]
)
def files_relpath(request: Any) -> List[Path]:
    params: List[str] = request.param
    return [Path(p) for p in params]


def test_find_config_file(tmp_path: Path, files_relpath: List[Path]) -> None:
    copy_config_file_to_dst("valid_sqlfmt_config.toml", tmp_path)

    files = [tmp_path / p for p in files_relpath]
    search_paths = _get_common_parents(files)
    assert tmp_path in search_paths
    config_path = _find_config_file(search_paths)
    assert config_path
    assert config_path == tmp_path / "pyproject.toml"


def test_find_config_file_no_file(tmp_path: Path, files_relpath: List[Path]) -> None:
    # don't copy the config file

    files = [tmp_path / p for p in files_relpath]
    search_paths = _get_common_parents(files)
    assert tmp_path in search_paths
    config_path = _find_config_file(search_paths)
    assert config_path is None


def test_find_config_file_not_in_tree(
    tmp_path: Path, files_relpath: List[Path]
) -> None:
    # put the config file outside the tree defined by files, so
    # we shouldn't find it
    config_dir = tmp_path / "config_dir"
    config_dir.mkdir()
    copy_config_file_to_dst("valid_sqlfmt_config.toml", config_dir)

    files = [tmp_path / p for p in files_relpath]
    search_paths = _get_common_parents(files)
    assert tmp_path in search_paths
    config_path = _find_config_file(search_paths)
    assert config_path is None


def test_load_config_from_path(tmp_path: Path) -> None:
    copy_config_file_to_dst("valid_sqlfmt_config.toml", tmp_path)
    config = _load_config_from_path(tmp_path / "pyproject.toml")
    assert config
    assert config["line_length"] == 100
    assert config["check"] is True
    assert config.get("name", "does not exist") == "does not exist"
    assert config.get("exclude_root", "does not exist") == "does not exist"


def test_load_config_from_path_minimal_config(tmp_path: Path) -> None:
    copy_config_file_to_dst("exclude_config.toml", tmp_path)
    config = _load_config_from_path(tmp_path / "pyproject.toml")
    assert config
    assert config["exclude"] == ["target/**/*", "dbt_packages/**/*"]
    assert config["exclude_root"] == tmp_path


def test_load_config_from_path_invalid_toml(tmp_path: Path) -> None:
    copy_config_file_to_dst("invalid_toml_config.toml", tmp_path)
    with pytest.raises(SqlfmtConfigError) as excinfo:
        _ = _load_config_from_path(tmp_path / "pyproject.toml")

    assert "Check for invalid TOML" in str(excinfo.value)


def test_load_config_from_path_invalid_key(tmp_path: Path) -> None:
    copy_config_file_to_dst("invalid_key_config.toml", tmp_path)
    with pytest.raises(SqlfmtConfigError) as excinfo:
        _ = _load_config_from_path(tmp_path / "pyproject.toml")

    assert "foo" in str(excinfo.value)


def test_load_config_from_path_dialect(tmp_path: Path) -> None:
    copy_config_file_to_dst("dialect_config.toml", tmp_path)
    config = _load_config_from_path(tmp_path / "pyproject.toml")
    assert config
    assert config["dialect_name"] == "clickhouse"


def test_load_config_from_path_dialect_name(tmp_path: Path) -> None:
    copy_config_file_to_dst("dialect_name_config.toml", tmp_path)
    config = _load_config_from_path(tmp_path / "pyproject.toml")
    assert config
    assert config["dialect_name"] == "clickhouse"


def test_load_config_from_missing_file(tmp_path: Path) -> None:
    copy_config_file_to_dst("invalid_toml_config.toml", tmp_path)
    with pytest.raises(SqlfmtConfigError) as excinfo:
        _ = _load_config_from_path(tmp_path / "not_pyproject.toml")

    assert "Error opening pyproject.toml" in str(excinfo.value)


def test_load_config_from_None(tmp_path: Path) -> None:
    copy_config_file_to_dst("invalid_toml_config.toml", tmp_path)
    config = _load_config_from_path(None)
    assert config == {}



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.clone import CLONE as CLONE
from sqlfmt.rules.common import (
    ALTER_DROP_FUNCTION,
    ALTER_WAREHOUSE,
    CREATE_CLONABLE,
    CREATE_FUNCTION,
    CREATE_WAREHOUSE,
    PRAGMA_SET_CALL,
    group,
)
from sqlfmt.rules.core import CORE as CORE
from sqlfmt.rules.function import FUNCTION as FUNCTION
from sqlfmt.rules.grant import GRANT as GRANT
from sqlfmt.rules.jinja import JINJA as JINJA  # noqa
from sqlfmt.rules.pragma import PRAGMA as PRAGMA
from sqlfmt.rules.unsupported import UNSUPPORTED as UNSUPPORTED
from sqlfmt.rules.warehouse import WAREHOUSE as WAREHOUSE
from sqlfmt.tokens import TokenType

MAIN = [
    *CORE,
    Rule(
        name="star_columns",
        priority=409,  # star is 410
        pattern=group(r"\*columns") + group(r"\(", r"$"),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.NAME),
    ),
    Rule(
        name="statement_start",
        priority=1000,
        pattern=group(r"case") + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.STATEMENT_START
            ),
        ),
    ),
    Rule(
        name="statement_end",
        priority=1010,
        pattern=group(r"end") + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.safe_add_node_to_buffer,
                token_type=TokenType.STATEMENT_END,
                fallback_token_type=TokenType.NAME,
            ),
        ),
    ),
    Rule(
        # a join's using word operator must be followed
        # by parens; otherwise, it's probably a
        # delete's USING, which is an unterminated
        # keyword
        name="join_using",
        priority=1049,
        pattern=group(r"using") + group(r"\s*\("),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.WORD_OPERATOR
            ),
        ),
    ),
    Rule(
        name="unterm_keyword",
        priority=1050,
        pattern=group(
            r"with(\s+recursive)?",
            (
                r"select(\s+(as\s+struct|as\s+value))?"
                r"(\s+(all|top\s+\d+|distinct))?"
                # select into is ddl that needs additional handling
                r"(?!\s+into)"
            ),
            r"delete\s+from",
            r"from",
            (
                r"((global|natural|asof|any)\s+)?"
                r"((inner|left|right|full|cross)\s+)?"
                r"((outer|semi|anti|any|all|asof|cross|positional|array|paste)\s+)?join"
            ),
            # this is the USING following DELETE, not the join operator
            # (see above)
            r"using",
            r"lateral\s+view(\s+outer)?",
            r"(pre)?where",
            r"group\s+by",
            r"cluster\s+by",
            r"distribute\s+by",
            r"sort\s+by",
            r"having",
            r"qualify",
            r"window",
            r"order\s+by",
            r"limit",
            r"fetch\s+(first|next)",
            r"for\s+(update|no\s+key\s+update|share|key\s+share)",
            r"when",
            r"then",
            r"else",
            r"partition\s+by",
            r"values",
            # in pg, RETURNING can be the last clause of
            # a DELETE statement
            r"returning",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
    Rule(
        # There are some function names in some dialects that are the same as
        # word operators in other dialects. Here we lex those as function
        # names IFF the name is immediately followed by a `(` (with no space
        # after the name. Otherwise they are lexed as word_operators by the
        # next rule.
        name="functions_that_overlap_with_word_operators",
        priority=1099,
        pattern=group(
            r"filter",
            r"isnull",
            r"(r|i)?like",
        )
        + group(r"\("),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(actions.add_node_to_buffer, token_type=TokenType.NAME),
        ),
    ),
    Rule(
        name="word_operator",
        priority=1100,
        pattern=group(
            r"as",
            r"(not\s+)?between",
            r"cube",
            r"(not\s+)?exists",
            r"filter",
            r"grouping sets",
            r"(global)?(not\s+)?in",
            r"interval",
            r"is(\s+not)?(\s+distinct\s+from)?",
            r"isnull",
            r"(not\s+)?i?like(\s+(any|all))?",
            r"over",
            r"(un)?pivot",
            r"notnull",
            r"(not\s+)?regexp",
            r"(not\s+)?rlike",
            r"rollup",
            r"some",
            r"(not\s+)?similar\s+to",
            r"tablesample",
            r"within\s+group",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.WORD_OPERATOR
            ),
        ),
    ),
    Rule(
        name="star_replace_exclude",
        priority=1101,
        pattern=group(
            r"exclude",
            r"replace",
        )
        + group(r"\s+\("),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.WORD_OPERATOR
            ),
        ),
    ),
    Rule(
        name="on",
        priority=1120,
        pattern=group(r"on") + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(actions.add_node_to_buffer, token_type=TokenType.ON),
        ),
    ),
    Rule(
        name="boolean_operator",
        priority=1200,
        pattern=group(
            r"and",
            r"or",
            r"not",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.BOOLEAN_OPERATOR
            ),
        ),
    ),
    Rule(
        name="frame_clause",
        priority=1305,
        pattern=group(r"(range|rows|groups)\s+")
        + group(r"(between\s+)?((unbounded|\d+)\s+(preceding|following)|current\s+row)")
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
    Rule(
        # BQ arrays use an offset(n) function for
        # indexing that we do not want to match. This
        # should only match the offset in limit ... offset,
        # which must be followed by a space
        name="offset_keyword",
        priority=1310,
        pattern=group(r"offset") + group(r"\s+", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
    Rule(
        name="set_operator",
        priority=1320,
        pattern=group(
            r"((inner|((full|left)(\s+outer)?))\s+)?"
            r"(union|intersect|except|minus)"
            r"(\s+(all|distinct))?"
            r"(\s+(by\s+name|((strict\s+)?corresponding(\s+by)?)))?",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=actions.handle_set_operator,
        ),
    ),
    Rule(
        name="explain",
        priority=2000,
        pattern=group(r"explain(\s+(analyze|verbose|using\s+(tabular|json|text)))?")
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_nonreserved_top_level_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
    Rule(
        name="pragma",
        priority=2005,
        pattern=group(PRAGMA_SET_CALL) + group(r"\W", r"$"),
        action=partial(
            actions.handle_nonreserved_top_level_keyword,
            action=partial(actions.lex_ruleset, new_ruleset=PRAGMA),
        ),
    ),
    Rule(
        name="grant",
        priority=2010,
        pattern=group(r"grant", r"revoke") + group(r"\W", r"$"),
        action=partial(
            actions.handle_nonreserved_top_level_keyword,
            action=partial(actions.lex_ruleset, new_ruleset=GRANT),
        ),
    ),
    Rule(
        name="create_clone",
        priority=2015,
        pattern=group(CREATE_CLONABLE + r"\s+.+?\s+clone") + group(r"\W", r"$"),
        action=partial(
            actions.handle_nonreserved_top_level_keyword,
            action=partial(
                actions.lex_ruleset,
                new_ruleset=CLONE,
            ),
        ),
    ),
    Rule(
        name="create_function",
        priority=2020,
        pattern=group(CREATE_FUNCTION, ALTER_DROP_FUNCTION) + group(r"\W", r"$"),
        action=partial(
            actions.handle_nonreserved_top_level_keyword,
            action=partial(
                actions.lex_ruleset,
                new_ruleset=FUNCTION,
            ),
        ),
    ),
    Rule(
        name="create_warehouse",
        priority=2030,
        pattern=group(
            CREATE_WAREHOUSE,
            ALTER_WAREHOUSE,
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_nonreserved_top_level_keyword,
            action=partial(
                actions.lex_ruleset,
                new_ruleset=WAREHOUSE,
            ),
        ),
    ),
    Rule(
        name="unsupported_ddl",
        priority=2999,
        pattern=group(
            r"alter",
            r"attach\s+rls\s+policy",
            r"cache\s+table",
            r"clear\s+cache",
            r"cluster",
            r"comment",
            r"copy",
            r"create",
            r"deallocate",
            r"declare",
            r"describe",
            r"desc\s+datashare",
            r"desc\s+identity\s+provider",
            r"delete",
            r"detach\s+rls\s+policy",
            r"discard",
            r"do",
            r"drop",
            r"execute",
            r"export",
            r"fetch",
            r"get",
            r"handler",
            r"import\s+foreign\s+schema",
            r"import\s+table",
            # snowflake: "insert into" or "insert overwrite into"
            # snowflake: has insert() function
            # spark: "insert overwrite" without the trailing "into"
            # redshift/pg: "insert into" only
            # bigquery: bare "insert" is okay
            r"insert(\s+overwrite)?(\s+into)?",
            r"list",
            r"lock",
            r"merge",
            r"move",
            # prepare transaction statements are simple enough
            # so we'll allow them
            r"prepare(?!\s+transaction)",
            r"put",
            r"reassign\s+owned",
            r"remove",
            r"rename\s+table",
            r"repair",
            r"security\s+label",
            r"select\s+into",
            r"truncate",
            r"unload",
            r"update",
            r"validate",
        )
        + r"(?!\()"
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_nonreserved_top_level_keyword,
            action=partial(
                actions.lex_ruleset,
                new_ruleset=UNSUPPORTED,
            ),
        ),
    ),
]
